#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#define MAX_SIZE 1000
#define MIN_SIZE 20
typedef enum transformation
{    ld, 
     lf, 
     lr, 
     li, 
     uls,
     nt
} transformation;
typedef struct header_detail
{
    char       variable_name[MIN_SIZE];
    char       lower_bound[MAX_SIZE];
    char       upper_bound[MAX_SIZE];
    char       cond[MIN_SIZE];
} header_detail;
typedef struct loop
{
    int           nest_count;
    int           line_count;
    char          loop_header[MIN_SIZE][MAX_SIZE];
    char          loop_body[MIN_SIZE][MAX_SIZE];
    header_detail hdl[MIN_SIZE];
    int           access_count_per_line[MIN_SIZE];
    char          access_exp[MIN_SIZE][MIN_SIZE][MAX_SIZE];
} loop;
typedef struct mapping
{
    char domain[MAX_SIZE][MAX_SIZE];
    char codomain[MAX_SIZE][MAX_SIZE];
    int  count;
} mapping;
typedef struct T_H_entry
{
    transformation t;
    int num_of_params;
    int param[MIN_SIZE];
} T_H_entry;

void print_mapping(mapping *M)
{
    int i;
    for(i = 0; i < M->count; i++)
    {
        printf("domain %s codomain %s\n", M->domain[i], M->codomain[i]);
    }
}

void print_loop(loop *L_s)
{
    int i, j;
    printf("\n-----LOOP-----\n");
    printf("Nest count %d\n", L_s->nest_count);
    printf("Line count %d\n", L_s->line_count);
    for(i = 0; i < L_s->nest_count; i++)
    {
        printf("%dth loop header %s\n", i, L_s->loop_header[i]);
    }
    for(i = 0; i < L_s->line_count; i++)
    {
        printf("%dth loop body %s\n", i, L_s->loop_body[i]);
    }
    printf("Header detail\n");
    for(i = 0; i < L_s->nest_count; i++)
    {
        printf("Variable name %s lower bound %s upper bound %s\n", L_s->hdl[i].variable_name, L_s->hdl[i].lower_bound, L_s->hdl[i].upper_bound);
    }
    for(i = 0; i < L_s->line_count; i++)
    {
	printf("Access count %d in %dth line\n", L_s->access_count_per_line[i], i);
        for(j = 0; j < L_s->access_count_per_line[i]; j++)
	{
	    printf("Exp %s ", L_s->access_exp[i][j]);
	}
	printf("\n");
    }
    printf("\n-----END LOOP-----\n");
}

char *replace_str(char *str, char *orig, char *rep)
{
    static char buffer[4096];
    char *p;
   
    if(!(p = strstr(str, orig)))
        return str;
		        
    strncpy(buffer, str, p-str);
    buffer[p-str] = ' ';
			        
    sprintf(buffer+(p-str), "%s%s", rep, p+strlen(orig));
    return buffer;
}


void replace_str_me(char *str, char *s1, char *s2)
{
    int i, j, k, l, n; 
    char str1[MIN_SIZE];

    printf("Old str %s s1 %s s2 %s\n", str, s1, s2);
    i = 0;
    while(strstr(str, s1))
    {
	sprintf(str1, "@%d", i);
        strcpy(str, replace_str(str, s1, str1));
	i++;
    }
    for(j = 0; j < i; j++)
    {
        sprintf(str1, "@%d", j);
	strcpy(str, replace_str(str, str1, s2));
    }
    printf("New str %s\n", str);
}



void readTwoloops(loop *L_s, loop *L_t, long offset_to_L_es, long offset_to_L_et)
{
    FILE *fp;
    char *Buf, *str, *token, *token1, *token2;
    int i;
    Buf = (char *)malloc(sizeof(char)*MAX_SIZE);
    str = (char *)malloc(sizeof(char)*MAX_SIZE);
    fp = fopen("loop", "r");
    for(i = 0; i < 2; i++)
    {    
	if(i == 0)
	{
	    fseek(fp, offset_to_L_es, SEEK_SET);
	    L_s->nest_count = 0;
	    L_s->line_count = 0;
	}
	else
	{
	    fseek(fp, offset_to_L_et, SEEK_SET);
	    L_t->nest_count = 0;
	    L_t->line_count = 0;
	}
        while(1)
        {
            fgets(Buf, MAX_SIZE, fp);
            if(Buf[0] == '-' && Buf[1] == '-')
                break;
	    else
            {
	        strcpy(str, Buf);
		token = strstr(str, "for");
		printf("Buf %s str %s token %s\n", Buf, str, token);
		if(token != NULL)
	        {
		    if(i == 0)
	            {
	                strcpy(L_s->loop_header[L_s->nest_count], Buf);
			printf("NCS %d content %s\n", L_s->nest_count, L_s->loop_header[L_s->nest_count]);
		        L_s->nest_count++;
		    }
		    else
		    {
	                strcpy(L_t->loop_header[L_t->nest_count], Buf);
			printf("NCT %d content %s\n", L_t->nest_count, L_t->loop_header[L_t->nest_count]);
			L_t->nest_count++;
		    }
		}
		else
		{
		    strcpy(str, Buf);
		    token1 = strstr(str, "{\n");
		    strcpy(str, Buf);
		    token2 = strstr(str, "}\n");
		    if(token1 != NULL || token2 != NULL)
		    {
			printf("Curly brace\n");
		        continue;
		    }
                    if(i == 0)
	            {
	                strcpy(L_s->loop_body[L_s->line_count], Buf);
		        L_s->line_count++;
		    }
		    else
	            {
	                strcpy(L_t->loop_body[L_t->line_count], Buf);
		        L_t->line_count++;
		    }
		}
            }
        }
	// printf("Nest count in %dth loop is %d\n", i+1, nest_count);
    }
    printf("L_s nest count %d line count %d L_t nest count %d line count %d\n", L_s->nest_count, L_s->line_count, L_t->nest_count, L_t->line_count);
    fclose(fp);
    free(Buf);
    free(str);
}

// L2 is copied to L1
void copy_loop(loop *L1, loop *L2)
{
    int i, j;

    L1->nest_count = L2->nest_count;
    L1->line_count = L2->line_count;
    for(j = 0; j < L2->nest_count; j++)
    {
        strcpy(L1->loop_header[j], L2->loop_header[j]);  
    }
    for(i = 0; i < L2->line_count; i++)
    {
        strcpy(L1->loop_body[i], L2->loop_body[i]);
	L1->access_count_per_line[i] = L2->access_count_per_line[i];
        for(j = 0; j < L2->access_count_per_line[i]; j++)
        {
            strcpy(L1->access_exp[i][j], L2->access_exp[i][j]);
        }
    }
    for(j = 0; j < L2->nest_count; j++)
    {
        strcpy(L1->hdl[j].variable_name, L2->hdl[j].variable_name);
        strcpy(L1->hdl[j].lower_bound, L2->hdl[j].lower_bound);
        strcpy(L1->hdl[j].upper_bound, L2->hdl[j].upper_bound);
    }
}

void filter_original_loop_body(loop *L1, loop *L2, loop *L_s, loop *L_t, int number_of_loops_src, int *number_of_statements_src, int number_of_loops_tgt, int *number_of_statements_tgt, transformation *ld_lf_flag)
{
    int i, j, k, count;

    *ld_lf_flag = nt;
    if(L1->line_count > L2->line_count)
    {
	*ld_lf_flag = ld;
	printf("HEADER1 ld\n");
        count = 0;
	for(i = 0; i < number_of_loops_tgt - 1; i++)
	{
	    count = count + number_of_statements_tgt[i];
	}
	L_s->nest_count = L1->nest_count;
	L_s->line_count = number_of_statements_tgt[number_of_loops_tgt - 1];
	for(j = 0; j < L1->nest_count; j++)
	{
	    strcpy(L_s->loop_header[j], L1->loop_header[j]);  
	}
	k = 0;
        for(i = count; i < count + number_of_statements_tgt[number_of_loops_tgt - 1]; i++, k++)
	{
	    strcpy(L_s->loop_body[k], L1->loop_body[i]);
	    for(j = 0; j < L1->access_count_per_line[i]; j++)
	    {
	        strcpy(L_s->access_exp[k][j], L1->access_exp[i][j]);
	    }
	}
        for(j = 0; j < L1->nest_count; j++)
        {
            strcpy(L_s->hdl[i].variable_name, L1->hdl[i].variable_name);
            strcpy(L_s->hdl[i].lower_bound, L1->hdl[i].lower_bound);
            strcpy(L_s->hdl[i].upper_bound, L1->hdl[i].upper_bound);
        }
	copy_loop(L_t, L2);
    }
    else if(L2->line_count > L1->line_count)
    {
	*ld_lf_flag = lf;
	printf("HEADER2 lf\n");
        count = 0;
	for(i = 0; i < number_of_loops_src - 1; i++)
	{
	    count = count + number_of_statements_src[i];
	}
	printf("Count %d\n", count);
	L_t->nest_count = L2->nest_count;
	L_t->line_count = number_of_statements_src[number_of_loops_src - 1];
	printf("Line count %d nest count %d\n", L_t->line_count, L_t->nest_count);
	for(j = 0; j < L2->nest_count; j++)
	{
	    strcpy(L_t->loop_header[j], L2->loop_header[j]);  
	}
	k = 0;
        for(i = count; i < count + number_of_statements_src[number_of_loops_src - 1]; i++, k++)
	{
	    strcpy(L_t->loop_body[k], L2->loop_body[i]);
	    for(j = 0; j < L2->access_count_per_line[i]; j++)
	    {
	        strcpy(L_t->access_exp[k][j], L2->access_exp[i][j]);
	    }
	}
        for(j = 0; j < L2->nest_count; j++)
        {
            strcpy(L_t->hdl[i].variable_name, L2->hdl[i].variable_name);
            strcpy(L_t->hdl[i].lower_bound, L2->hdl[i].lower_bound);
            strcpy(L_t->hdl[i].upper_bound, L2->hdl[i].upper_bound);
        }
        copy_loop(L_s, L1);
        print_loop(L_t);
    }
    else
    {
	printf("HEADER3\n");
        copy_loop(L_s, L1);
	copy_loop(L_t, L2);
    }
    print_loop(L_s);
    print_loop(L_t);
}

void find_loop_pair_in_error(loop *L_s, loop *L_t, int *number_of_loops_src, int *number_of_statements_per_loop_src, int *number_of_loops_tgt, int *number_of_statements_per_loop_tgt, transformation *ld_lf_flag)
{
    int i, j, k1, k2, in_loop_flag;
    FILE *fp;
    char *Buf, *str, *token;
    long offset, offset_to_L_es, offset_to_L_et;
    //int  number_of_loops_src, number_of_statements_per_loop_src[MIN_SIZE], number_of_loops_tgt, number_of_statements_per_loop_tgt[MIN_SIZE];
    loop L1, L2;
    
    fp = fopen("loop", "r");
    in_loop_flag = 0;
    *number_of_loops_src = 0; 
    *number_of_loops_tgt = 0;
    k1 = k2 = -1;
    Buf = (char *)malloc(sizeof(char)*MAX_SIZE);
    str = (char *)malloc(sizeof(char)*MAX_SIZE);
    for(i = 0; i < 2; i++)
    {
        while(1)
        {
	    offset = ftell(fp);
            fgets(Buf, MAX_SIZE, fp);
            if(Buf[0] == '-' && Buf[1] == '-')
                break;
	    else
            {
	        strcpy(str, Buf);
		token = strstr(str, "for");
		printf("Buf %s str %s token %s\n", Buf, str, token);
		if(token != NULL)
	        {
		    if(in_loop_flag == 0)
		    {
		        in_loop_flag = 1;
			if(i == 0)
			{
			    offset_to_L_es = offset;
			    *number_of_loops_src = *number_of_loops_src + 1;
			    k1++;
			}
			else
			{
			    offset_to_L_et = offset;
			    *number_of_loops_tgt = *number_of_loops_tgt + 1;
			    k2++;
			}
		    }
		}
		else
		{
		    if(in_loop_flag == 1)
		    {
		        in_loop_flag = 0;
			if(i == 0)
			{
			    number_of_statements_per_loop_src[k1] = 0;
			}
			else
			{
			    number_of_statements_per_loop_tgt[k2] = 0;
			}
		    }
		    strcpy(str, Buf);
		    token = strstr(str, ";");
		    if(token != NULL)
		    {        
			if(i == 0)
			{
		            number_of_statements_per_loop_src[k1] = number_of_statements_per_loop_src[k1] + 1;
			}
			else
			{
			    number_of_statements_per_loop_tgt[k2] = number_of_statements_per_loop_tgt[k2] + 1;
			}
		    }
		}
            }
        }
	// printf("Nest count in %dth loop is %d\n", i+1, nest_count);
    }
    printf("Number of loops src %d tgt %d\n", *number_of_loops_src, *number_of_loops_tgt);
    for(i = 0; i <= k1; i++)
    {
        printf("Number of statements src %d tgt %d\n", number_of_statements_per_loop_src[i], number_of_statements_per_loop_tgt[i]);
    }
    readTwoloops(&L1, &L2, offset_to_L_es, offset_to_L_et);
    filter_original_loop_body(&L1, &L2, L_s, L_t, *number_of_loops_src, number_of_statements_per_loop_src, *number_of_loops_tgt, number_of_statements_per_loop_tgt, ld_lf_flag);
    print_loop(&L1);
    print_loop(&L2);
    free(Buf);
    free(str);
    fclose(fp);
}


void readHeaderDetail(loop *L)
{
    int i, j, k, flag;
    char *token, *token1, *token2, *str, *var_decl, *var_cond, *lower_bound;


    str = (char *)malloc(sizeof(char)*MAX_SIZE);
    var_decl = (char *)malloc(sizeof(char)*MAX_SIZE);
    var_cond = (char *)malloc(sizeof(char)*MAX_SIZE);
    lower_bound = (char *)malloc(sizeof(char)*MAX_SIZE);
    printf("Nest count %d\n", L->nest_count);
    for(i = 0; i < L->nest_count; i++)
    {
       strcpy(str, L->loop_header[i]);
       token = strstr(str, "for");
       token = token+4;
       strcpy(str, token);
       token = strtok(str, ";");
       // extract variable declaration
       if(token!= NULL)
       {
           // printf ("DECL %s\n", token);
	   strcpy(var_decl, token);
	   token = strtok (NULL, ";");
       }
       // extract variable condition
       if(token!= NULL)
       {
           // printf ("COND %s\n", token);
           strcpy(var_cond, token);
           token = strtok (NULL, ";");
       }
       for(j = 0; var_decl[j] != '='; j++)
       {
           L->hdl[i].variable_name[j] = var_decl[j];	       
       }
       L->hdl[i].variable_name[j] = '\0';
       j++;
       for(k = 0; j != ';'; j++, k++)
       {
           L->hdl[i].lower_bound[k] = var_decl[j];
           //lower_bound[k] = var_decl[j];
       }
       L->hdl[i].lower_bound[k] = '\0';
       //lower_bound[k] = '\0';
       //L->hdl[i].lower_bound = atoi(lower_bound);
       // printf("lower bound %s\n", L->hdl[i].lower_bound);
       // printf("Var cond %s\n", var_cond);

       strcpy(str, var_cond);
       token = strstr(str, "<=");
       if(token != NULL)
       {
           token = strstr(str, "<=");
           strcpy(L->hdl[i].cond, "<=");
	   token = token + 2; 
           for(k = 0; token[k] != '\0'; k++)
	   {
	       L->hdl[i].upper_bound[k] = token[k];
	   }
           L->hdl[i].upper_bound[k] = '\0';
       }
       else
       {
           token = strstr(str, ">=");
           if(token != NULL)
	   {
               strcpy(L->hdl[i].cond, ">=");
	       token = token + 2; 
               for(k = 0; token[k] != '\0'; k++)
	       {
	           L->hdl[i].upper_bound[k] = token[k];
	       }
               L->hdl[i].upper_bound[k] = '\0';
           }
	   else
           {
               token = strstr(str, "<");
               if(token != NULL)
	       {
	           strcpy(L->hdl[i].cond, "<");
	           token = token + 1; 
                   for(k = 0; token[k] != '\0'; k++)
	           {
	               L->hdl[i].upper_bound[k] = token[k];
	           }
                   L->hdl[i].upper_bound[k] = '\0';
	       }
	       else
	       {
	           token = strstr(str, ">");
		   if(token != NULL)
		   {       
	               strcpy(L->hdl[i].cond, ">");
	               token = token + 1; 
                       for(k = 0; token[k] != '\0'; k++)
	               {
	                   L->hdl[i].upper_bound[k] = token[k];
	               }
                       L->hdl[i].upper_bound[k] = '\0';
		   }
		   else
		   {
		   }
	       }
           }
       }
       // printf("Upper bound %s\n", L->hdl[i].upper_bound);
       // printf("cond %s\n", L->hdl[i].cond);
    }
    free(str);
    free(var_decl);
    free(var_cond);
    free(lower_bound);
}

void readBodyDetail(loop *L)
{
    int i, j, n, k, index, flag;

    // printf("Line count %d\n", L->line_count);
    for(i = 0; i < L->line_count; i++)
    {
        k = 0;
	n = 0;
	flag = 0;
	// printf("LB %s\n", L->loop_body[i]);
        for(j = 0; j < strlen(L->loop_body[i]) - 1; j++)
	{
            if(L->loop_body[i][j] == '[' && flag == 0)
            {
	        flag = 1;
		index = 0;
		//printf("%c", L_s->loop_body[i][j]);
            }

	    if(flag == 1)
	    {
		// printf("%c", L->loop_body[i][j]);
		L->access_exp[i][k][index] = L->loop_body[i][j];
		index++;
            }

	    if(L->loop_body[i][j] == ']' && (L->loop_body[i][j+1] == ' ' || L->loop_body[i][j+1] == ')' || L->loop_body[i][j+1] == ';') && (flag == 1))
            {
		L->access_exp[i][k][index] = '\0';
	        flag = 0;
                n++;
		k++;
		index = 0;
	    }
	}
	// printf("Number of accesses %d\n", n);
        L->access_count_per_line[i] = k;
    }
    /*for(j = 0; j < L->line_count; j++)
    {
        for(i = 0; i < k; i++)
        {
            printf("Exp %s\n", L->access_exp[j][i]);
        }
    }
    printf("Access count %d\n", k);*/
}

void construct_M_H(mapping *M_H, loop *L_s, loop *L_t) 
{
    int i;
    M_H->count = 0;
    // construct M_H
    for(i = 0; i < L_s->nest_count; i++)
    {
        strcpy(M_H->domain[i], L_s->hdl[i].variable_name);
	strcpy(M_H->codomain[i], L_t->hdl[i].variable_name);
	M_H->count++;
    }
}

void take_majority(mapping *M2, int count, mapping *M_B, loop *L_s)
{
    int i, j, *check, *count_of_occurrences, *max_occurrences, *max_occurrences_index;

    check = (int*)malloc(sizeof(int)*MAX_SIZE);
    count_of_occurrences = (int*)malloc(sizeof(int)*MAX_SIZE);
    max_occurrences = (int*)malloc(sizeof(int)*MIN_SIZE);
    max_occurrences_index = (int*)malloc(sizeof(int)*MIN_SIZE);

    for(i = 0; i < count; i++)
    {
        check[i] = 0; 
	count_of_occurrences[i] = 0;
	printf("M2 %s->%s\n", M2->domain[i], M2->codomain[i]);
    }

    for(i = 0; i < count - 1; i++)
    {
        check[i] = 1;
	count_of_occurrences[i] = 1;
        for(j = i + 1; j < count; j++)
        {
            if(!strcmp(M2->domain[i], M2->domain[j]) && !strcmp(M2->codomain[i], M2->codomain[j]))
	    {
	        check[j] = 1;
		count_of_occurrences[i] = count_of_occurrences[i] + 1;
	    }
        }
    }
    /*for(i = 0; i < count; i++)
    {
        printf("Count of occurrences %d\n", count_of_occurrences[i]);
    }*/

    for(i = 0; i < L_s->nest_count; i++)
    {
        max_occurrences[i] = 0;
    }

    printf("Printing LOOP\n");
    print_loop(L_s);
    for(i = 0; i < L_s->nest_count; i++)
    {
        for(j = 0; j < count; j++)
	{
	    //printf("Variable name %s domain name %s !strcmp(L_s->hdl[i].variable_name, M2->domain[j]) %d strlen(L_s->hdl[i].variable_name) %d strlen(M2->domain[j]) %d\n", L_s->hdl[i].variable_name, M2->domain[j], !strcmp(L_s->hdl[i].variable_name, M2->domain[j]), strlen(L_s->hdl[i].variable_name), strlen(M2->domain[j]));
            if(!strcmp(L_s->hdl[i].variable_name, M2->domain[j]))
	    {
		// printf("Ekhane aschhe na\n");
                if(count_of_occurrences[j] >= max_occurrences[i])
		{
	            max_occurrences[i] = count_of_occurrences[j];
		    max_occurrences_index[i] = j;
		}
	    }
	}
	printf("Var name %s Max occurrences %d max_occurrences_index %d mapping %s -> %s\n", L_s->hdl[i].variable_name, max_occurrences[i], max_occurrences_index[i], M2->domain[max_occurrences_index[i]], M2->codomain[max_occurrences_index[i]]);
        strcpy(M_B->domain[i], M2->domain[max_occurrences_index[i]]);
        strcpy(M_B->codomain[i], M2->codomain[max_occurrences_index[i]]);
	printf("M_B %s->%s\n", M_B->domain[i], M_B->codomain[i]);
    }
    M_B->count = L_s->nest_count;
    free(check);
    free(count_of_occurrences);
    free(max_occurrences);
    free(max_occurrences_index);
}

void construct_M_B(mapping *M_B, loop *L_s, loop *L_t)
{
    int i, k, j, lesser;
    mapping *M1, *M2;
    char *token1, *token2, *str1, *str2, *str3;


    // construct M_B
    M1 = (mapping *)malloc(sizeof(mapping));
    M2 = (mapping *)malloc(sizeof(mapping));
    str1 = (char *)malloc(sizeof(char)*MAX_SIZE);
    str2 = (char *)malloc(sizeof(char)*MAX_SIZE);
    str3 = (char *)malloc(sizeof(char)*MAX_SIZE);
    /*for(j = 0; j < L_s->line_count; j++)
    {
        for(i = 0; i < L_s->access_count_per_line[j]; i++)
        {
            printf("bsExp %s\n", L_s->access_exp[j][i]);
        }
    }
    for(j = 0; j < L_t->line_count; j++)
    {
        for(i = 0; i < L_t->access_count_per_line[j]; i++)
        {
            printf("btExp %s\n", L_t->access_exp[j][i]);
        }
    }*/
    M1->count = 0;
    if(L_s->line_count <= L_t->line_count)
    {
        lesser = 1;
    }
    else
    {
        lesser = 2;
    }
    if(lesser == 1)
    {
	for(j = 0; j < L_s->line_count; j++)
	{
            for(i = 0; i < L_s->access_count_per_line[j]; i++)
	    {
	        strcpy(M1->domain[i], L_s->access_exp[j][i]);
	        strcpy(M1->codomain[i], L_t->access_exp[j][i]);
	        M1->count++;
	    }
	}
    }
    else
    {
        for(j = 0; j < L_t->line_count; j++)
	{
            for(i = 0; i < L_t->access_count_per_line[j]; i++)
	    {
	        strcpy(M1->domain[i], L_s->access_exp[j][i]);
	        strcpy(M1->codomain[i], L_t->access_exp[j][i]);
	        M1->count++;
	    }
	}
    }
    for(i = 0; i < M1->count; i++)
    {
        printf("M_B : %s -> %s\n", M1->domain[i], M1->codomain[i]);
    }

    j = 0;
    for(i = 0; i < M1->count; i++)
    {
        strcpy(str1, M1->domain[i]);
	token1 = strtok(str1, "[]");
	while(token1 != NULL) 
	{
	    strcpy(M2->domain[j], token1);
	    // printf("TOKEN1 %s\n", token1);
	    j++;
	    token1 = strtok(NULL, "[]");
	}
    }
    j = 0;
    for(i = 0; i < M1->count; i++)
    {
        strcpy(str2, M1->codomain[i]);
	token2 = strtok(str2, "[]");
	while(token2 != NULL) 
	{
	    strcpy(M2->codomain[j], token2);
	    // printf("TOKEN2 %s\n", token2);
	    j++;
	    token2 = strtok(NULL, "[]");
	}
    }
    printf("MAPPING\n");
    print_mapping(M2);
    take_majority(M2, j, M_B, L_s);
    free(str1);
    free(str2);
    free(str3);
    free(M1);
    free(M2);
}

//void set_params_ld_lf(loop *L, T_H_entry *t_h, long offset_to_remaining_loops)
/*void set_params_ld_lf(loop *L, T_H_entry *t_h)
{
    int  i, in_loop_flag;
    char *str, *token, *Buf, count_st_in_loop_in_error, count_st_in_loop;
    FILE *fp;

    // printf("PRINTING FOR DEBUGGING OFFSET %ld\n", offset_to_remaining_loops);
    str = (char *)malloc(sizeof(char)*MAX_SIZE);
    Buf = (char *)malloc(sizeof(char)*MAX_SIZE);
    count_st_in_loop_in_error = 0;
    t_h->num_of_params = 0;
    for(i = 0; i < L->line_count; i++)
    {
        strcpy(str, L->loop_body[i]);
	token = strtok(str, ";");
        if(token != NULL)
	{
	    count_st_in_loop_in_error++;
	}
    }
    t_h->param[t_h->num_of_params] = count_st_in_loop_in_error;

    in_loop_flag = 0;
    fp = fopen("loop", "r");
    //fseek(fp, offset_to_remaining_loops, SEEK_SET);
    t_h->param[t_h->num_of_params] = 0;
    
    while(1)
    {
        fgets(Buf, MAX_SIZE, fp);
	printf("Buf %s\n", Buf);
	strcpy(str, Buf);
        if(str[0] == '-' && str[1] == '-')
            break;
        else
        {
	    token = strstr(str, "for");
            if(token != NULL)
	    {
	        if(in_loop_flag == 0)
		{
                    t_h->num_of_params++;
                    t_h->param[t_h->num_of_params] = 0;
		}
                in_loop_flag = 1;
	    }
	    else  
	    {
	        if(in_loop_flag == 1)
	        {
                    in_loop_flag = 0;
	        }
		strcpy(str, Buf);
		token = strtok(str, ";");
		if(token != NULL)
	        {
	            // Counting number of statements in a loop 
                    t_h->param[t_h->num_of_params]++;
		}

	    }
        }
    }
    t_h->num_of_params++;
    printf("number of params for ld %d\n", t_h->num_of_params);
    free(Buf);
    free(str);
    fclose(fp);
}*/
void ascertain_presence_of_uls_li_lr(loop *L_s, loop *L_t, T_H_entry *t_h, mapping *M_B)
{
    int i, j, k, l, flag;
    char var[MIN_SIZE], ub[MIN_SIZE], lb[MIN_SIZE], str[MIN_SIZE], sf[MIN_SIZE], *ret, str1[MIN_SIZE], str2[MIN_SIZE];

    printf("Ascertain presence of ULS, LR and LI\n");
    flag = 0;
    for(i = 0; i < L_s->nest_count; i++)
    {
        strcpy(str, M_B->codomain[i]);
        for(j = 0; j < L_s->nest_count; j++)
        {
	    for(k = 0; k < L_s->nest_count; k++)
	    {
                sprintf(str1, "%s-%s", L_t->hdl[j].variable_name, L_t->hdl[k].variable_name);		
                if(!strcmp(str, str1))
                {
		    printf("IL %d JL %d KL %d\n", i, j, k);
                    if(i != j)
		    {
		       t_h[li].t = li;
		       t_h[li].num_of_params = 2;
		       printf("LI ? 3 i %d j %d\n", i, j);
		       if(i < j)
		       {
		           t_h[li].param[0] = i;
		           t_h[li].param[1] = j;
		       }
		       else
		       {
		           t_h[li].param[0] = j;
		           t_h[li].param[1] = i;
		       }
		    }

		    for(l = 0; l < L_s->nest_count; l++)
		    {
			if(!strcmp(M_B->codomain[l], L_t->hdl[k].variable_name))
			{
                            t_h[uls].t = uls;
                            t_h[uls].num_of_params = 3;
	                    t_h[uls].param[0] = i;
	                    t_h[uls].param[1] = l;
                            t_h[uls].param[2] = 1;
			    flag = 1;
			    break;
			}
                    }
                    if(l == L_s->nest_count && flag == 1)
		    {
                        t_h[uls].t = uls;
                        t_h[uls].num_of_params = 3;
	                t_h[uls].param[0] = i;
                        t_h[uls].param[1] = j;
		        t_h[uls].param[2] = 1;
		        t_h[li].num_of_params = 2;
                        if(i != j)
		        {
		           t_h[li].t = li;
		           t_h[li].num_of_params = 2;
		           printf("LI ? 4 i %d j %d\n", i, j);
		           if(i < j)
		           {
		               t_h[li].param[0] = i;
		               t_h[li].param[1] = j;
		           }
		           else
		           {
		               t_h[li].param[0] = j;
		               t_h[li].param[1] = i;
		           }
		        }
		    }
		}
	        else
	        {
                    sprintf(str1, "%s+%s", L_t->hdl[j].variable_name, L_t->hdl[k].variable_name);		
	            if(!strcmp(str, str1))
	            {
		       printf("IL %d JL %d KL %d\n", i, j, k);
		       sprintf(str2, "-%s", L_t->hdl[k].variable_name);
		       for(l = 0; l < L_s->nest_count; l++)
		       {
		           if(!strcmp(M_B->codomain[l], str2))
                           {
                               t_h[uls].t = uls;
                               t_h[uls].num_of_params = 3;
                               t_h[uls].param[0] = i;
        		       t_h[uls].param[1] = l;
			       t_h[uls].param[2] = 1;
			       t_h[lr].t = lr;
			       t_h[lr].num_of_params = 1;
			       t_h[lr].param[0] = l;
			       flag = 1;
			       break;
			   }
			}
                        if(l == L_s->nest_count && flag == 1)
		        {
                            t_h[uls].t = uls;
                            t_h[uls].num_of_params = 3;
	                    t_h[uls].param[0] = i;
                            t_h[uls].param[1] = j;
		            t_h[uls].param[2] = 1;
                            t_h[li].t = li;
		            t_h[li].num_of_params = 2;
		            printf("LI ? 5 i %d j %d\n", i, j);
                            if(i < j)
		            {
		                t_h[li].param[0] = i;
		                t_h[li].param[1] = j;
		            }
		            else
		            {
		                t_h[li].param[0] = j;
		                t_h[li].param[1] = i;
		            }
		        }
	            }
	            else
	            {
                        sprintf(str1, "-%s-%s", L_t->hdl[j].variable_name, L_t->hdl[k].variable_name);		
	                if(!strcmp(str, str1))
	                {
			    sprintf(str2, "%s", L_t->hdl[k].variable_name);
			    // printf("Eikhane\n");
		            // printf("IL %d JL %d KL %d\n", i, j, k);
		            for(l = 0; l < L_s->nest_count; l++)
		            {
			        // printf("Eikhane M_B->codomain[l] %s str2 %s\n", M_B->codomain[l], str2);
		      	        if(!strcmp(M_B->codomain[l], str2))
			        {
			            // printf("DEBUG\n");
                                    t_h[uls].t = uls;
                                    t_h[uls].num_of_params = 3;
	                            t_h[uls].param[0] = i;
			            t_h[uls].param[1] = l;
			            t_h[uls].param[2] = 1;
                                    t_h[lr].t = lr;
			            t_h[lr].num_of_params = 1;
			            t_h[lr].param[0] = i;
			            flag = 1;
			            break;
			        }
			    }
                            if(l == L_s->nest_count && flag == 1)
		            {
                                t_h[uls].t = uls;
                                t_h[uls].num_of_params = 3;
	                        t_h[uls].param[0] = i;
                                t_h[uls].param[1] = j;
		                t_h[uls].param[2] = 1;
                                t_h[li].t = li;
		                t_h[li].num_of_params = 2;
		                // printf("LI ? 6 i %d j %d\n", i, j);
                                if(i < j)
		                {
		                    t_h[li].param[0] = i;
		                    t_h[li].param[1] = j;
		                }
		                else
		                {
		                    t_h[li].param[0] = j;
		                    t_h[li].param[1] = i;
		                }
		            }
	                }
		    }
		}
	    }
	}
    }
    if(t_h[uls].num_of_params > 0)
    {
	// printf("Khoma t_h[uls].param[0] %d t_h[uls].param[1] %d\n", t_h[uls].param[0], t_h[uls].param[1]);
        if(t_h[uls].param[0] > t_h[uls].param[1])
	{
	    i = t_h[uls].param[0];
	    t_h[uls].param[0] = t_h[uls].param[1];
	    t_h[uls].param[1] = i;
	}
	else
	{
	}
    }
    // printf("Hence, t_h[uls].param[0] %d t_h[uls].param[1] %d\n", t_h[uls].param[0], t_h[uls].param[1]);
}

void ascertain_presence_of_li(loop *L_s, loop *L_t, T_H_entry *t_h, mapping *M_B)
{
    int i, j, k, l, flag;
    char var[MIN_SIZE], ub[MIN_SIZE], lb[MIN_SIZE], str[MIN_SIZE], sf[MIN_SIZE], *ret, str1[MIN_SIZE], str2[MIN_SIZE];
    
    // printf("Ascertain presence of LI\n");
    for(i = 0; i < L_s->nest_count; i++)
    {
	/*for(k = 0; k < L_s->nest_count; k++)
	{
            if(!strcmp(M_B->domain[i], L_s->hdl[k].variable_name))
	    {
                break;
	    }
	}*/

       for(k = 0; k < L_s->nest_count; k++)
       {
            if(!strcmp(M_B->codomain[i], L_t->hdl[k].variable_name))
	    {
                break;
            }
       }
       if(k == L_s->nest_count)
       {
           break;
       }

       if(k != i)
       {
            // m1 and m2 as per rule 3 are found
    	    t_h[li].t = li;
    	    t_h[li].num_of_params = 2;
	    printf("LI ? 2 i %d j %d\n", k,  j);
	    if(k < i)
	    {
    	        t_h[li].param[0] = k;
    	        t_h[li].param[1] = i;
	    }
	    else
	    {
    	        t_h[li].param[0] = i;
    	        t_h[li].param[1] = k;
	    }
        }
    }
}

void ascertain_presence_of_lr_li(loop *L_s, loop *L_t, T_H_entry *t_h, mapping *M_B)
{
    int i, j, k, l, flag;
    char var[MIN_SIZE], ub[MIN_SIZE], lb[MIN_SIZE], str[MIN_SIZE], sf[MIN_SIZE], *ret, str1[MIN_SIZE], str2[MIN_SIZE];

    printf("Ascertain presence of LR and LI\n");
    for(i = 0; i < L_s->nest_count; i++)
    {
	for(j = 0; j < L_t->nest_count; j++)
	{
            sprintf(var, "-%s", L_t->hdl[j].variable_name);
            if(!strcmp(M_B->codomain[i], var))
	    {
	        t_h[lr].t = lr;
	        t_h[lr].num_of_params = 1;
	        t_h[lr].param[0] = i;
		if(i != j)
		{
		    t_h[li].t = li;
	    	    t_h[li].num_of_params = 2;
		    if(i < j)
		    {
	    	        t_h[li].param[0] = i;
	    	        t_h[li].param[1] = j;
		    }
		    else
		    {
	    	        t_h[li].param[0] = j;
	    	        t_h[li].param[1] = i;
		    }
		}
	        break;
	    }
	}
    }
}

void construct_T_H(mapping *M_H, mapping *M_B, loop *L_s, loop *L_t, T_H_entry *t_h, int number_of_loops_src, int *number_of_statements_per_loop_src, int number_of_loops_tgt, int *number_of_statements_per_loop_tgt, int *sequence, transformation ld_lf_flag)
{
    int i, j, k, l, flag;
    char var[MIN_SIZE], ub[MIN_SIZE], lb[MIN_SIZE], str[MIN_SIZE], sf[MIN_SIZE], *ret, str1[MIN_SIZE], str2[MIN_SIZE];

    printf("LD LF FLAG %d\n", ld_lf_flag);

    /* Initialize */
    for(i = 0; i < 5; i++)
    {
        t_h[i].t = nt;
	t_h[i].num_of_params = 0;
    }

    /* Presence of ld or lf and presence of lr */
    if(ld_lf_flag == ld)
    {
        t_h[ld].t = ld;
	printf("T_H %d\n", t_h[ld].t);
	t_h[ld].num_of_params = number_of_loops_tgt;
	//set_params_ld_lf(L_t, &t_h[0], offset_to_remaining_loops);
	//set_params_ld_lf(L_t, &t_h[0]);
	for(i = 0; i < number_of_loops_src; i++)
	{
	    t_h[ld].param[i] = number_of_statements_per_loop_tgt[i];
	}
    }
    else if(ld_lf_flag == lf)
    {
        t_h[lf].t = lf;
	printf("T_H %d\n", t_h[lf].t);
	//set_params_ld_lf(L_s, &t_h[1], offset_to_remaining_loops);
	//set_params_ld_lf(L_s, &t_h[1]);
	t_h[lf].num_of_params = number_of_loops_src;
        for(i = 0; i < number_of_loops_src; i++)
	{
	    t_h[lf].param[i] = number_of_statements_per_loop_src[i];
	    printf("PARAMS %d\n", t_h[lf].param[i]);
	}
    }
    else
    {
    }

    /* Presence of li*/
    for(i = 0; i < L_s->nest_count; i++)
    {
        printf("M_H : domain %s codomain %s\n", M_H->domain[i], M_H->codomain[i]);
        printf("M_B : domain %s codomain %s\n", M_B->domain[i], M_B->codomain[i]);
    }
    ascertain_presence_of_li(L_s, L_t, t_h, M_B);
    printf("TAU_H1\n");
    for(i = 0; i < 5; i++)
    {
        printf("%d ", t_h[i].t);
    }

    /* Presence of lr and li */
    ascertain_presence_of_lr_li(L_s, L_t, t_h, M_B);
    printf("TAU_H2\n");
    for(i = 0; i < 5; i++)
    {
        printf("%d ", t_h[i].t);
    }

    /* Presence of uls, li and lr */
    ascertain_presence_of_uls_li_lr(L_s, L_t, t_h, M_B);
    printf("TAU_H3\n");
    for(i = 0; i < 5; i++)
    {
        printf("Transformation %d ", t_h[i].t);
	for(j = 0; j < t_h[i].num_of_params; j++)
	{
	    printf("param %d value %d\n", j, t_h[i].param[j]);
	}
	printf("\n");
    }
    printf("Source loop\n");
    print_loop(L_s);
    printf("Transformed loop\n");
    print_loop(L_t);
    printf("\n");
}

int apply_ld(loop *L_s, loop *L_is, T_H_entry T_H[])
{
    //int i, j, k, l, count;
    copy_loop(L_is, L_s);
    return 1;
}

int apply_lf(loop *L_s, loop *L_is, T_H_entry T_H[])
{
    /*int i, j, k, l;

    printf("IN APPLYING LF\n");
    for(j = 0; j < L_s->nest_count; j++)
    {
        strcpy(L_is->loop_header[j], L_s->loop_header[j]);
        strcpy(L_is->hdl[j].variable_name, L_s->hdl[j].variable_name);
        strcpy(L_is->hdl[j].lower_bound, L_s->hdl[j].lower_bound);
        strcpy(L_is->hdl[j].upper_bound, L_s->hdl[j].upper_bound);
        strcpy(L_is->hdl[j].cond, L_s->hdl[j].cond);
    }
    for(j = 0; j < T_H[lf].param[T_H[lf].num_of_params - 1]; j++)
    {
        strcpy(L_is->loop_body[j], L_s->loop_body[j]);
        for(k = 0; k < L_s->access_count_per_line[j]; k++)
	{
	    strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
	} 
    }
    L_is->nest_count = L_s->nest_count;
    L_is->line_count = T_H[lf].param[T_H[lf].num_of_params - 1];*/
    copy_loop(L_is, L_s);
    return 1;
}


int apply_lr(loop *L_s, loop *L_is, T_H_entry T_H[], int uls_flag, int li_flag)
{
    int j, k, param0;
    char str[MIN_SIZE], neg_str[MIN_SIZE], pos_str[MIN_SIZE], pos_str1[MIN_SIZE], pos_uls_str[MIN_SIZE], neg_uls_str[MIN_SIZE];

    L_is->nest_count = L_s->nest_count;
    L_is->line_count = L_s->line_count;
    param0 = T_H[lr].param[0];
    if(li_flag == 1)
    {
        if(T_H[lr].param[0] == T_H[li].param[0])
	{
	    param0 = T_H[li].param[1];
	}
	else if(T_H[lr].param[0] == T_H[li].param[1])
	{
	    param0 = T_H[li].param[0];
	}
    }
    printf("param0 %d\n", param0);

    for(j = 0; j < L_s->nest_count; j++)
    {
        strcpy(L_is->loop_header[j], L_s->loop_header[j]);
        strcpy(L_is->hdl[j].variable_name, L_s->hdl[j].variable_name);
        strcpy(L_is->hdl[j].lower_bound, L_s->hdl[j].lower_bound);
        strcpy(L_is->hdl[j].upper_bound, L_s->hdl[j].upper_bound);
        strcpy(L_is->hdl[j].cond, L_s->hdl[j].cond);
    }
    sprintf(neg_str, "-%s", L_is->hdl[param0].variable_name);
    sprintf(pos_str, "%s", L_is->hdl[param0].variable_name);
    sprintf(pos_str1, "+%s", L_is->hdl[param0].variable_name);
    for(j = 0; j < L_s->line_count; j++)
    {
	L_is->access_count_per_line[j] = L_s->access_count_per_line[j];
        strcpy(L_is->loop_body[j], L_s->loop_body[j]);
	for(k = 0; k < L_s->access_count_per_line[j]; k++)
	{
            //strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
            strcpy(str, L_s->access_exp[j][k]);
	    if(!uls_flag)
	    {
	        if(strstr(str, neg_str))
	        {
		    if(!strcmp(str, neg_str))
		    {
	                printf("Inverse application\n");
	                // strcpy(L_is->access_exp[j][k], replace_str(L_s->access_exp[j][k], neg_str, pos_str));
			replace_str_me(L_s->access_exp[j][k], neg_str, pos_str);
	                strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
		    }
		    else
		    {
		        //strcpy(L_is->access_exp[j][k], replace_str(L_s->access_exp[j][k], neg_str, pos_str1));
			replace_str_me(L_is->access_exp[j][k], neg_str, pos_str1);
		        strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
		    }
	        }
	        else if(strstr(str, pos_str))
	        {
	            printf("Forward application\n");
	            // strcpy(L_is->access_exp[j][k], replace_str(L_s->access_exp[j][k], pos_str, neg_str));
		    replace_str_me(L_s->access_exp[j][k], pos_str, neg_str);
	            strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
	        }
	    }
            else
	    {
	        //sprintf(pos_uls_str, "%s-%d.%s", L_is->hdl[param1].variable_name, T_H[uls].param[2], L_is->hdl[param0].variable_name);
	        //sprintf(neg_uls_str, "-(%s-%d.%s)", L_is->hdl[param1].variable_name, T_H[uls].param[2], L_is->hdl[param0].variable_name);
                if(strstr(str, neg_str))
                {
		    if(!strcmp(str, neg_str))
		    {
	                // strcpy(L_is->access_exp[j][k], replace_str(L_s->access_exp[j][k], neg_str, pos_str));
			replace_str_me(L_s->access_exp[j][k], neg_str, pos_str);
	                strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
		    }
		    else
		    {
	                //strcpy(L_is->access_exp[j][k], replace_str(L_s->access_exp[j][k], neg_str, pos_str1));
			replace_str_me(L_s->access_exp[j][k], neg_str, pos_str1);
	                strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
		    }
		}
		else if(strstr(str, pos_str))
		{
		    replace_str_me(L_s->access_exp[j][k], pos_str, neg_str);
                    strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
		}
	    }
	}
    }
    strcpy(str, L_is->hdl[param0].lower_bound);
    if(strlen(L_is->hdl[param0].upper_bound) == 1 || (strlen(L_is->hdl[param0].upper_bound) == 2 && atoi(L_is->hdl[param0].upper_bound) != 0))
    {
        sprintf(L_is->hdl[param0].lower_bound, "-%s", L_is->hdl[param0].upper_bound);
    }
    else
    {
        sprintf(L_is->hdl[param0].lower_bound, "-(%s)", L_is->hdl[param0].upper_bound);
    }

    if(strlen(str) == 1 || (strlen(str) == 2 && atoi(str) != 0))
    {
        sprintf(L_is->hdl[param0].upper_bound, "-%s", str); 
    }
    else
    {
        sprintf(L_is->hdl[param0].upper_bound, "-(%s)", str); 
    }
    print_loop(L_is);
    return 1;
}

void tokenize(char str[], char c1, char c2, char token_array[][MIN_SIZE], int *token_count, char delim_array[][MIN_SIZE], int *delim_count)
{
    int i, j = 0;

    for(i = 0; i < strlen(str); i++)
    {
        if(str[i] == c1)
	{
	   delim_array[*delim_count][0] = c1; 
	   delim_array[*delim_count][1] = '\0';
	   *delim_count = *delim_count + 1;
	   token_array[*token_count][j] = '\0';
	   *token_count = *token_count + 1;
	   j = 0;
	}
        else if(str[i] == c2)
	{
	    delim_array[*delim_count][0] = c2;
	    delim_array[*delim_count][1] = '\0';
	    *delim_count = *delim_count + 1;
	    token_array[*token_count][j] = '\0';
	    *token_count = *token_count + 1;
	    j = 0;
	}
	else
	{
	    token_array[*token_count][j] = str[i];
	    j++;
	}
    }
    for(i = 0; i < *token_count; i++)
    {
        printf("TOKEN %s\n", token_array[i]);
    }
    for(i = 0; i < *delim_count; i++)
    {
        printf("DELIM %s\n", delim_array[i]);
    }

}

void apply_FM_method(loop *L_s, loop *L_is, T_H_entry T_H[])
{
    int  i, j, k, k1, k2, token_count, delim_count;
    char vn[MIN_SIZE], ub[MIN_SIZE], lb[MIN_SIZE], str[MIN_SIZE], token_array[MIN_SIZE][MIN_SIZE], delim_array[MIN_SIZE][MIN_SIZE];

    k1 = T_H[li].param[0];
    k2 = T_H[li].param[1];
    printf("Interchanged k1 %d k2 %d\n", k1, k2);
    strcpy(L_is->hdl[k1].variable_name, L_s->hdl[k1].variable_name);
    strcpy(L_is->hdl[k1].lower_bound, L_s->hdl[k1].lower_bound);
    strcpy(L_is->hdl[k1].upper_bound, L_s->hdl[k1].upper_bound);
    strcpy(L_is->hdl[k2].variable_name, L_s->hdl[k2].variable_name);
    strcpy(L_is->hdl[k2].lower_bound, L_s->hdl[k2].lower_bound);
    strcpy(L_is->hdl[k2].upper_bound, L_s->hdl[k2].upper_bound);


    strcpy(L_is->hdl[k1].variable_name, L_s->hdl[k2].variable_name);
    strcpy(L_is->hdl[k2].variable_name, L_s->hdl[k1].variable_name);
    token_count = delim_count = 0;
    // printf("To tokenize %s", L_is->hdl[k2].lower_bound);
    tokenize(L_is->hdl[k2].lower_bound, '+', '-', token_array, &token_count, delim_array, &delim_count);		    
    j = 0;
    strcpy(lb, "");
    strcat(lb, token_array[0]);
    for(i = 1; i < token_count; i++)
    {
        if(!strstr(token_array[i], "."))
	{
	    strcat(lb, delim_array[j]);
	    j++;
	    strcat(lb, token_array[i]);
	}
    }
    printf("To tokenize %s", L_is->hdl[k2].upper_bound);
    token_count = delim_count = 0;
    printf("To tokenize %s", L_is->hdl[k2].lower_bound);
    tokenize(L_is->hdl[k2].upper_bound, '+', '-', token_array, &token_count, delim_array, &delim_count);		    
    j = 0;
    strcpy(ub, "");
    strcat(ub, token_array[0]);
    for(i = 1; i < token_count; i++)
    {
        if(!strstr(token_array[i], "."))
	{
	    strcat(ub, delim_array[j]);
	    j++;
	    strcat(ub, token_array[i]);
	}
    }
    printf("Found out LB %s UB %s k1 lb %s ub %s k2 lb %s ub %s\n", lb, ub, L_is->hdl[k1].lower_bound, L_s->hdl[k1].upper_bound, L_is->hdl[k2].lower_bound, L_s->hdl[k2].upper_bound);
    k = atoi(lb)+atoi(L_s->hdl[k1].lower_bound);
    sprintf(L_is->hdl[k1].lower_bound, "%d", k);
    k = atoi(ub)+atoi(L_s->hdl[k1].upper_bound);
    sprintf(L_is->hdl[k1].upper_bound, "%d", k);
    sprintf(L_is->hdl[k2].lower_bound, "max(%s,%s-%s)", L_s->hdl[k1].lower_bound, L_is->hdl[k1].variable_name, ub);
    sprintf(L_is->hdl[k2].upper_bound, "min(%s,%s-%s)", L_s->hdl[k1].upper_bound, L_is->hdl[k1].variable_name, lb);
}

int apply_li(loop *L_s, loop *L_is, T_H_entry T_H[])
{
    int  i, j, k, k1, k2;
    char header[MAX_SIZE], vn[MIN_SIZE], lb[MAX_SIZE], ub[MAX_SIZE], *token1, *token2;

    L_is->nest_count = L_s->nest_count;
    L_is->line_count = L_s->line_count;
    for(j = 0; j < L_s->nest_count; j++)
    {
        strcpy(L_is->loop_header[j], L_s->loop_header[j]);
        strcpy(L_is->hdl[j].variable_name, L_s->hdl[j].variable_name);
        strcpy(L_is->hdl[j].lower_bound, L_s->hdl[j].lower_bound);
        strcpy(L_is->hdl[j].upper_bound, L_s->hdl[j].upper_bound);
        strcpy(L_is->hdl[j].cond, L_s->hdl[j].cond);
    }

    for(j = 0; j < L_s->line_count; j++)
    {
        strcpy(L_is->loop_body[j], L_s->loop_body[j]);
        L_is->access_count_per_line[j] = L_s->access_count_per_line[j];
	for(k = 0; k < L_s->access_count_per_line[j]; k++)
	{
            strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
	}
    }

    // Check if Fourier-Motzkin's method is applicable
    k1 = T_H[li].param[0];
    strcpy(lb, L_is->hdl[k1].lower_bound);
    strcpy(ub, L_is->hdl[k1].upper_bound);
    for(i = 0; i < k1; i++)
    {
        token1 = strstr(lb, L_is->hdl[i].variable_name);
        token2 = strstr(ub, L_is->hdl[i].variable_name);
	if(token1 != NULL || token2 != NULL)
	    break;
    }
    k2 = T_H[li].param[1];
    strcpy(lb, L_is->hdl[k2].lower_bound);
    strcpy(ub, L_is->hdl[k2].upper_bound);
    for(j = 0; j < k2; j++)
    {
        token1 = strstr(lb, L_is->hdl[j].variable_name);
        token2 = strstr(ub, L_is->hdl[j].variable_name);
        if(token1 != NULL || token2 != NULL)
            break;
    }
    if(i == k1 && j == k2)
    {
        // Fourier-Motzkin's method is not applicable
        strcpy(header, L_is->loop_header[T_H[li].param[0]]);
        strcpy(vn, L_is->hdl[T_H[li].param[0]].variable_name);
        strcpy(lb, L_is->hdl[T_H[li].param[0]].lower_bound);
        strcpy(ub, L_is->hdl[T_H[li].param[0]].upper_bound);
        strcpy(L_is->loop_header[T_H[li].param[0]], L_is->loop_header[T_H[li].param[1]]);
        strcpy(L_is->hdl[T_H[li].param[0]].variable_name, L_is->hdl[T_H[li].param[1]].variable_name);
        strcpy(L_is->hdl[T_H[li].param[0]].lower_bound, L_is->hdl[T_H[li].param[1]].lower_bound);
        strcpy(L_is->hdl[T_H[li].param[0]].upper_bound, L_is->hdl[T_H[li].param[1]].upper_bound);
        
        strcpy(L_is->loop_header[T_H[li].param[1]], header);
        strcpy(L_is->hdl[T_H[li].param[1]].variable_name, vn);
        strcpy(L_is->hdl[T_H[li].param[1]].lower_bound, lb);
        strcpy(L_is->hdl[T_H[li].param[1]].upper_bound, ub);
    }
    else
    {
        // Fourier-Motzkin's method
	printf("IN FM method\n");
	apply_FM_method(L_s, L_is, T_H);
    }
    return 1;
}

void apply_uls_header(loop *L_s, loop *L_is, T_H_entry T_H[], int param0, int param1)
{
    int j;
    char str[MIN_SIZE];

    
    printf("ULS Header in T_H 0 %d 1 %d 2 %d\n", param0, param1, T_H[uls].param[2]);
    L_is->nest_count = L_s->nest_count;
    for(j = 0; j < L_s->nest_count; j++)
    {
        strcpy(L_is->loop_header[j], L_s->loop_header[j]);
        strcpy(L_is->hdl[j].variable_name, L_s->hdl[j].variable_name);
        strcpy(L_is->hdl[j].lower_bound, L_s->hdl[j].lower_bound);
        strcpy(L_is->hdl[j].upper_bound, L_s->hdl[j].upper_bound);
        strcpy(L_is->hdl[j].cond, L_s->hdl[j].cond);
    }
    printf("Originally0 lb %s ub %s varname %s\n", L_is->hdl[param0].lower_bound, L_is->hdl[param0].upper_bound, L_is->hdl[param0].variable_name);
    printf("Originally1 lb %s ub %s varname %s\n", L_is->hdl[param1].lower_bound, L_is->hdl[param1].upper_bound, L_is->hdl[param1].variable_name);
    strcpy(str, L_is->hdl[param1].lower_bound);
    //sprintf(L_is->hdl[T_H[uls].param[1]].lower_bound, "%s+%d.%s", str, T_H[uls].param[2], L_is->hdl[T_H[uls].param[0]].variable_name);
    sprintf(L_is->hdl[param1].lower_bound, "%s+%s", str, L_is->hdl[param0].variable_name);
    strcpy(str, L_is->hdl[param1].upper_bound);
    //sprintf(L_is->hdl[T_H[uls].param[1]].upper_bound, "%s+%d.%s", str, T_H[uls].param[2], L_is->hdl[T_H[uls].param[0]].variable_name);
    sprintf(L_is->hdl[param1].upper_bound, "%s+%s", str, L_is->hdl[param0].variable_name);
    printf("Finally0 lb %s varname %s\n", L_is->hdl[param0].lower_bound, L_is->hdl[param0].variable_name);
    printf("Finally1 lb %s varname %s\n", L_is->hdl[param1].lower_bound, L_is->hdl[param1].variable_name);
}

int find_pos(char *str1, char *str2)
{
    int i, j, k;
    for(i = 0; i < strlen(str1); i++)
    {
	k = i;
        for(j = 0; j < strlen(str2); )
	{
	    if(str1[k] != str2[j])
	    {
                break;
	    }
	    else
	    {
	        k++; j++;
	    }
	}
	if(j == strlen(str2))
	{
	    return i;
	}
    }
}

void apply_uls_body(loop *L_s, loop *L_is, T_H_entry T_H[], int param0, int param1)
{
    int  i, j, k, l, pos;
    char str[MAX_SIZE], str1[MAX_SIZE];

    // printf("apply_uls_body IN T_H[uls].param[0] %d T_H[uls].param[1] %d\n", param0, param1);
    L_is->line_count = L_s->line_count;
    for(i = 0; i < L_s->line_count; i++)
    {
        L_is->access_count_per_line[i] = L_s->access_count_per_line[i];
	// printf("Access count %d in %dth line\n", L_is->access_count_per_line[i], i);
	for(j = 0; j < L_s->access_count_per_line[i]; j++)
	{
	    strcpy(L_is->access_exp[i][j], L_s->access_exp[i][j]);
	    pos = find_pos(L_is->access_exp[i][j], L_is->hdl[param1].variable_name);
	    // printf("Access exp %s pos %d\n", L_is->access_exp[i][j], pos);
	    //sprintf(str, "%s-%d.%s", L_is->hdl[T_H[uls].param[1]].variable_name, T_H[uls].param[2], L_is->hdl[T_H[uls].param[0]].variable_name);
	    sprintf(str, "%s-%s", L_is->hdl[param1].variable_name, L_is->hdl[param0].variable_name);
	    for(l = 0, k = pos+1; k < strlen(L_is->access_exp[i][j]); l++, k++)
            {
	        str1[l] = L_is->access_exp[i][j][k];
	    }
	    str1[l] = '\0';
	    for(l = pos, k = 0; k < strlen(str); l++, k++)
	    {
	        L_is->access_exp[i][j][l] = str[k];
	    }
	    for(k = 0; k < strlen(str1); k++, l++)
	    {
	        L_is->access_exp[i][j][l] = str1[k];
	    }
	    L_is->access_exp[i][j][l] = '\0';
        }	
    }
    // printf("apply_uls_body OUT\n");
}

int apply_uls(loop *L_s, loop *L_is, T_H_entry T_H[], int li_flag)
{
    int temp, param0, param1;

    param0 = T_H[uls].param[0];
    param1 = T_H[uls].param[1];
    if(li_flag == 1)
    {
        if(T_H[uls].param[0] == T_H[li].param[0])
	{
	    param0 = T_H[li].param[1];
	}	
	if(T_H[uls].param[1] == T_H[li].param[0])
	{
	    param1 = T_H[li].param[1];
	}
        if(T_H[uls].param[0] == T_H[li].param[1])
	{
	    param0 = T_H[li].param[0];
	}	
	if(T_H[uls].param[1] == T_H[li].param[1])
	{
	    param1 = T_H[li].param[0];
	}
    }
    if(param0 > param1)
    {
        temp = param0;
	param0 = param1;
	param1 = temp;
    }
    printf("param0 %d param1 %d\n", param0, param1);
    apply_uls_header(L_s, L_is, T_H, param0, param1);
    apply_uls_body(L_s, L_is, T_H, param0, param1);
    return 1;
}

int construct_forward_sequence(loop *L_s, loop *L_t, loop *L_i, T_H_entry T_H[], int perm[][MIN_SIZE], int n, int length_of_sequence)
{
    int i, j, k, count, ret_val, pos, uls_flag = 0, li_flag = 0;
    loop L_is[5];
    
    L_is[0] = *L_s;
    ret_val = 1;
    pos = length_of_sequence;
    for(i = 0; i < pos; i++)
    {
	printf("Perm[n][i] %d\n", perm[n][i]);
        if(perm[n][i] == ld)
	{
	    printf("In LD\n");
            ret_val = ret_val * apply_ld(&L_is[i], &L_is[i+1], T_H);
	    // print_loop(&L_is[i]);
   	}
	else if(perm[n][i] == lf)
	{
            ret_val = ret_val * apply_lf(&L_is[i], &L_is[i+1], T_H);
	}
	else if(perm[n][i] == lr)
	{
	    printf("In LR\n");
            ret_val = ret_val * apply_lr(&L_is[i], &L_is[i+1], T_H, uls_flag, li_flag);
            print_loop(&L_is[i+1]);
	}
	else if(perm[n][i] == li)
	{
	    li_flag = 1;
            ret_val = ret_val * apply_li(&L_is[i], &L_is[i+1], T_H);
	}
	else if(perm[n][i] == uls)
	{
	    ret_val = ret_val * apply_uls(&L_is[i], &L_is[i+1], T_H, li_flag);
	    uls_flag = 1;
	}
        printf("After applying %d FS, GENERATED LOOP\n", perm[n][i]);
        print_loop(&L_is[i+1]);
    }
    //perm_check_list[perm_cll] = ret;
    copy_loop(L_i, &L_is[pos]); 
    // printf("OUT\n");
    return ret_val;
}

int inverse_apply_ld(loop *L_t, loop *L_os, T_H_entry T_H[])
{
/*    int i, j, k, l, count;

    count = 0;
    for(j = 0; j < T_H[ld].num_of_params - 1; j++)
    {
        count += T_H[ld].param[j];
    }
    for(j = 0; j < L_t->nest_count; j++)
    {
        strcpy(L_os->loop_header[j], L_t->loop_header[j]);
        strcpy(L_os->hdl[j].variable_name, L_t->hdl[j].variable_name);
        strcpy(L_os->hdl[j].lower_bound, L_t->hdl[j].lower_bound);
        strcpy(L_os->hdl[j].upper_bound, L_t->hdl[j].upper_bound);
        strcpy(L_os->hdl[j].cond, L_t->hdl[j].cond);
    }
    l = 0; 
    for(j = count; j < count + T_H[ld].param[T_H[ld].num_of_params - 1]; j++, k++)
    {
        strcpy(L_os->loop_body[k], L_t->loop_body[j]);
        for(k = 0; k < L_t->access_count_per_line[j]; k++)
	{
	    strcpy(L_os->access_exp[l][k], L_t->access_exp[j][k]);
	    l++;
	}
    }
    for(j = 0; j < L_t->nest_count; j++)
    {
        L_os->hdl[j] = L_t->hdl[j];
    }
    L_os->nest_count = L_t->nest_count;
    L_os->line_count = T_H[ld].param[T_H[ld].num_of_params - 1]; */
    copy_loop(L_os, L_t);
    return 1;
}

int inverse_apply_lf(loop *L_s, loop *L_is, T_H_entry T_H[])
{
/*    int i, j, k, l, count;

    printf("IN APPLYING LF\n");
    count = 0;
    for(j = 0; j < T_H[lf].num_of_params - 1; j++)
    {
        count += T_H[lf].param[j];
    }
    printf("Count lf %d\n", count);
    for(j = 0; j < L_s->nest_count; j++)
    {
        strcpy(L_is->loop_header[j], L_s->loop_header[j]);
        strcpy(L_is->hdl[j].variable_name, L_s->hdl[j].variable_name);
        strcpy(L_is->hdl[j].lower_bound, L_s->hdl[j].lower_bound);
        strcpy(L_is->hdl[j].upper_bound, L_s->hdl[j].upper_bound);
        strcpy(L_is->hdl[j].cond, L_s->hdl[j].cond);
    }
    l = 0;
    for(j = count; j < count + T_H[lf].param[T_H[lf].num_of_params - 1]; j++, l++)
    {
        strcpy(L_is->loop_body[l], L_s->loop_body[j]);
        for(k = 0; k < L_s->access_count_per_line[j]; k++)
	{
	    strcpy(L_is->access_exp[l][k], L_s->access_exp[j][k]);
	} 
    }
    L_is->nest_count = L_s->nest_count;
    L_is->line_count = T_H[lf].param[T_H[lf].num_of_params - 1];*/
    copy_loop(L_is, L_s);
    return 1;
}

char *strrev(char *str)
{
    char *p1, *p2;
    if(! str || ! *str)
        return str;

    for (p1 = str, p2 = str + strlen(str) - 1; p2 > p1; ++p1, --p2)
    {
        *p1 ^= *p2;
        *p2 ^= *p1;
        *p1 ^= *p2;
    }

    return str;
}

int inverse_apply_lr(loop *L_s, loop *L_is, T_H_entry T_H[], int uls_flag, int li_flag)
{
    return apply_lr(L_s, L_is, T_H, uls_flag, li_flag);
    /*int j, k;
    char str[MIN_SIZE];

    L_is->nest_count = L_s->nest_count;
    L_is->line_count = L_s->line_count;
    for(j = 0; j < L_s->nest_count; j++)
    {
        strcpy(L_is->loop_header[j], L_s->loop_header[j]);
        strcpy(L_is->hdl[j].variable_name, L_s->hdl[j].variable_name);
        strcpy(L_is->hdl[j].lower_bound, L_s->hdl[j].lower_bound);
        strcpy(L_is->hdl[j].upper_bound, L_s->hdl[j].upper_bound);
        strcpy(L_is->hdl[j].cond, L_s->hdl[j].cond);

    }
    sprintf(str, "-%s", L_is->hdl[T_H[lr].param[0]].variable_name);
    for(j = 0; j < L_s->line_count; j++)
    {
        strcpy(L_is->loop_body[j], L_s->loop_body[j]);
	for(k = 0; k < L_s->access_count_per_line[j]; k++)
	{
            strcpy(L_is->access_exp[j][k], L_s->access_exp[j][k]);
	    strcpy(L_is->access_exp[j][k], replace_str(L_is->access_exp[j][k], str, L_is->hdl[T_H[lr].param[0]].variable_name));
	}
    }
    // sprintf(L_is->hdl[T_H[lr].param[0]].lower_bound, "%s", strrev(strncpy(L_is->hdl[T_H[lr].param[0]].upper_bound, strrev(L_is->hdl[T_H[lr].param[0]].upper_bound), strlen(L_is->hdl[T_H[lr].param[0]].upper_bound - 1))));
    // sprintf(L_is->hdl[T_H[lr].param[0]].upper_bound, "%s", strrev(strncpy(L_is->hdl[T_H[lr].param[0]].lower_bound, strrev(L_is->hdl[T_H[lr].param[0]].lower_bound), strlen(L_is->hdl[T_H[lr].param[0]].lower_bound - 1))));
    strcpy(str, L_is->hdl[T_H[lr].param[0]].lower_bound);
    sprintf(L_is->hdl[T_H[lr].param[0]].lower_bound, "%d", -1*atoi(L_is->hdl[T_H[lr].param[0]].upper_bound));
    sprintf(L_is->hdl[T_H[lr].param[0]].upper_bound, "%d", -1*atoi(str));*/
}

int inverse_apply_FM_method(loop *L_s, loop *L_os, T_H_entry T_H[])
{
    int  k, k1, k2;
    char vn[MIN_SIZE], str[MAX_SIZE], *token;

    /* Parsing and populating lower and upper bound of the interchanged loops from the string of the form [max(l1, i2 - h2), min(h1, i2 - l2)]*/
    k1 = T_H[li].param[0];
    k2 = T_H[li].param[1];
    strcpy(vn, L_os->hdl[k1].variable_name);
    strcpy(L_os->hdl[k1].variable_name, L_os->hdl[k2].variable_name);
    strcpy(L_os->hdl[k2].variable_name, vn);
    strcpy(str, L_os->hdl[k2].lower_bound);
    if(strstr(str, "max"))
    {
        strcpy(str, L_os->hdl[k2].lower_bound);
        token = strtok(str, "(,)");
        token = strtok(NULL, "(,)");
        strcpy(L_os->hdl[k1].lower_bound, token);
        token = strtok(NULL, "(,)");
        strcpy(str, token);
        token = strtok(str, "-+");
        token = strtok(NULL, "-+");
        strcpy(L_os->hdl[k2].upper_bound, token);
    }
    else
    {
        return 0;
    }

    strcpy(str, L_os->hdl[k2].upper_bound);
    if(strstr(str, "min"))
    {
        strcpy(str, L_os->hdl[k2].upper_bound);
        token = strtok(str, "(,)");
        token = strtok(NULL, "(,)");
        strcpy(L_os->hdl[k1].upper_bound, token);
        token = strtok(NULL, "(,)");
        strcpy(str, token);
        token = strtok(str, "-+");
        token = strtok(NULL, "-+");
        strcpy(L_os->hdl[k2].lower_bound, token);
    }
    else
    {
        return 0;
    }
    return 1;
    /* k = atoi(L_is->hdl[k1].lower_bound)+atoi(L_is->hdl[k2].lower_bound);
    sprintf(L_is->hdl[k1].lower_bound, "%d", k);
    k = atoi(L_is->hdl[k1].upper_bound)+atoi(L_is->hdl[k2].upper_bound);
    sprintf(L_is->hdl[k1].upper_bound, "%d", k);
    sprintf(L_is->hdl[k2].lower_bound, "max(%s, %s - %s)", L_is->hdl[k1].lower_bound, L_is->hdl[k1].variable_name, L_is->hdl[k2].upper_bound);
    sprintf(L_is->hdl[k2].upper_bound, "min(%s, %s - %s)", L_is->hdl[k1].upper_bound, L_is->hdl[k1].variable_name, L_is->hdl[k2].lower_bound); */
    //sprintf(L_is->hdl[k1].lower_bound, "%d",
}

int inverse_apply_li(loop *L_s, loop *L_os, T_H_entry T_H[])
{
    int  i, j, k, k1, k2;
    char header[MAX_SIZE], vn[MIN_SIZE], lb[MAX_SIZE], ub[MAX_SIZE], *token1, *token2;

    L_os->nest_count = L_s->nest_count;
    L_os->line_count = L_s->line_count;
    for(j = 0; j < L_s->nest_count; j++)
    {
        strcpy(L_os->loop_header[j], L_s->loop_header[j]);
        strcpy(L_os->hdl[j].variable_name, L_s->hdl[j].variable_name);
        strcpy(L_os->hdl[j].lower_bound, L_s->hdl[j].lower_bound);
        strcpy(L_os->hdl[j].upper_bound, L_s->hdl[j].upper_bound);
        strcpy(L_os->hdl[j].cond, L_s->hdl[j].cond);
    }

    for(j = 0; j < L_s->line_count; j++)
    {
        strcpy(L_os->loop_body[j], L_s->loop_body[j]);
	for(k = 0; k < L_s->access_count_per_line[j]; k++)
	{
            strcpy(L_os->access_exp[j][k], L_s->access_exp[j][k]);
	}
    }

    // Check if Fourier-Motzkin's method is applicable
    k1 = T_H[li].param[0];
    strcpy(lb, L_os->hdl[k1].lower_bound);
    strcpy(ub, L_os->hdl[k1].upper_bound);
    for(i = 0; i < k1; i++)
    {
        token1 = strstr(lb, L_os->hdl[i].variable_name);
        token2 = strstr(ub, L_os->hdl[i].variable_name);
	if(token1 != NULL || token2 != NULL)
	    break;
    }
    k2 = T_H[li].param[1];
    strcpy(lb, L_os->hdl[k2].lower_bound);
    strcpy(ub, L_os->hdl[k2].upper_bound);
    for(j = 0; j < k2; j++)
    {
        token1 = strstr(lb, L_os->hdl[j].variable_name);
        token2 = strstr(ub, L_os->hdl[j].variable_name);
        if(token1 != NULL || token2 != NULL)
            break;
    }
    if(i == k1 && j == k2)
    {
        // Fourier-Motzkin's method is not applicable
        strcpy(header, L_os->loop_header[T_H[li].param[0]]);
        strcpy(vn, L_os->hdl[T_H[li].param[0]].variable_name);
        strcpy(lb, L_os->hdl[T_H[li].param[0]].lower_bound);
        strcpy(ub, L_os->hdl[T_H[li].param[0]].upper_bound);

        strcpy(L_os->loop_header[T_H[li].param[0]], L_os->loop_header[T_H[li].param[1]]);
        strcpy(L_os->hdl[T_H[li].param[0]].variable_name, L_os->hdl[T_H[li].param[1]].variable_name);
        strcpy(L_os->hdl[T_H[li].param[0]].lower_bound, L_os->hdl[T_H[li].param[1]].lower_bound);
        strcpy(L_os->hdl[T_H[li].param[0]].upper_bound, L_os->hdl[T_H[li].param[1]].upper_bound);
        strcpy(L_os->loop_header[T_H[li].param[1]], header);
        strcpy(L_os->hdl[T_H[li].param[1]].variable_name, vn);
        strcpy(L_os->hdl[T_H[li].param[1]].lower_bound, lb);
        strcpy(L_os->hdl[T_H[li].param[1]].upper_bound, ub);
    }
    else
    {
        // Inverse application of Fourier-Motzkin's method
	return inverse_apply_FM_method(L_s, L_os, T_H);
    }
    return 1;
}

int inverse_apply_uls_header(loop *L_s, loop *L_os, T_H_entry T_H[])
{
    int i, j;
    char str[MAX_SIZE], str1[MAX_SIZE], str2[MAX_SIZE], *token, *token1;

    printf("INVERSE APPLY ULS HEADER\n");
    sprintf(str, "%s+", L_os->hdl[T_H[uls].param[0]].variable_name);
    if(strstr(L_os->hdl[T_H[uls].param[1]].lower_bound, str))
    {
        strcpy(str1, L_os->hdl[T_H[uls].param[1]].lower_bound);
	j = 0;
	for(i = strlen(str); i < strlen(str1); i++, j++)
	{
	    L_os->hdl[T_H[uls].param[1]].lower_bound[j] = L_os->hdl[T_H[uls].param[1]].lower_bound[i];
	}
	L_os->hdl[T_H[uls].param[1]].lower_bound[j] = '\0';
	strcpy(str1, L_os->hdl[T_H[uls].param[1]].upper_bound);
	j = 0;
	for(i = strlen(str); i < strlen(str1); i++, j++)
	{
	    L_os->hdl[T_H[uls].param[1]].upper_bound[j] = L_os->hdl[T_H[uls].param[1]].upper_bound[i];
	}
	L_os->hdl[T_H[uls].param[1]].upper_bound[j] = '\0';
    }
    return 1;
}

void find_pos_char_count(char *access_exp, char c, int pos_arr[], int *pos_count)
{ 
    int i;
    for(i = 0; i < strlen(access_exp); i++)
    {
        if(access_exp[i] == c)
	{
	    pos_arr[*pos_count] = i;
	    *pos_count = *pos_count + 1;
	}
    }  
}



int inverse_apply_uls_body(loop *L_s, loop *L_os, T_H_entry T_H[])
{
    int i, j, k, l, m, n, p, delim_index;
    char str[MAX_SIZE], str1[MAX_SIZE], str2[MAX_SIZE], exp[MAX_SIZE], access_exp[MIN_SIZE][MIN_SIZE], exp_arr[MAX_SIZE], *token;
    int dot_pos[MIN_SIZE], dot_count, token_count, delim_count;
    char c1, c2, token_array[MIN_SIZE][MIN_SIZE], delim_array[MIN_SIZE][MIN_SIZE];
    
    printf("INVERSE APPLY ULS BODY\n");
    L_os->line_count = L_s->line_count;
    L_os->nest_count = L_s->nest_count;
    for(i = 0; i < L_os->line_count; i++)
    {
	L_os->access_count_per_line[i] = L_s->access_count_per_line[i];
	strcpy(L_os->loop_body[i], L_s->loop_body[i]);
        for(j = 0; j < L_s->access_count_per_line[i]; j++)
	{
            printf("Access exp %s\n", L_s->access_exp[i][j]);
            strcpy(str, L_s->access_exp[i][j]);
	    l = 0;
            token = strtok(str, "[]");
            while(token != NULL)
	    {
		strcpy(access_exp[l], token);
		token = strtok(NULL, "[]");
		l++;
	    }

	    m = 0;
	    for(k = 0; k < l; k++)
	    {
		dot_count = 0;
                find_pos_char_count(access_exp[k], '.', dot_pos, &dot_count);
		exp_arr[m] = '[';
		m++;
		if(dot_count == 0)
		{
		    for(n = 0; n < strlen(access_exp[k]); n++, m++)
		    {
		        exp_arr[m] = access_exp[k][n];
		    }
		}
		else
		{
		    strcpy(str, access_exp[k]);
		    token_count = delim_count = 0;
                    tokenize(str, '+', '-', token_array, &token_count, delim_array, &delim_count);		    
		    for(p = 0; p < strlen(token_array[0]); p++, m++)
		    {
	                exp_arr[m] = token_array[0][p];
		    }
		    delim_index = 0;
		    for(n = 1; n < token_count; )
		    {
	                if(strstr(token_array[n], "."))
			{
			    delim_index++;
			    n++;
			}
			else
		        {
			    for(p = 0; p < strlen(delim_array[delim_index]); p++, m++)
			    {
		                exp_arr[m] = delim_array[delim_index][p];
			    }
			    for(p = 0; p < strlen(token_array[n]); p++, m++)
			    {
                                exp_arr[m] = token_array[n][p];      
			    }
			    delim_index++;
			    n++;
			}
		    }
		}
		exp_arr[m] = ']';
		m++;
	    }	
	    exp_arr[m] = '\0';
	    strcpy(L_os->access_exp[i][j], exp_arr);
	}
    }
    return 1;
}

void inverse_apply_uls(loop *L_s, loop *L_os, T_H_entry T_H[])
{
    inverse_apply_uls_header(L_s, L_os, T_H);
    /* Body is still remaining */
    inverse_apply_uls_body(L_s, L_os, T_H);
}

int construct_backward_sequence(loop *L_t, loop *L_ib, T_H_entry T_H[], int perm[][MIN_SIZE], int n, int length, int pos)
{
    int i, j, k, count, ret_val, uls_flag = 0, li_flag = 0;
    loop L_os[5], *L;
    
    printf("In construct backward sequence; pos %d n %d\n", pos, n);
    L_os[length] = *L_t;
    for(i = length; i > pos; i--)
    {
        if(perm[n][i-1] == ld)
	{
            printf("In inverse apply ld");
            ret_val = inverse_apply_ld(&L_os[i], &L_os[i-1], T_H);
            print_loop(&L_os[i-1]);
   	}
	else if(perm[n][i-1] == lf)
	{
            inverse_apply_lf(&L_os[i], &L_os[i-1], T_H);
	}
	else if(perm[n][i-1] == lr)
	{
            printf("In inverse apply lr");
            inverse_apply_lr(&L_os[i], &L_os[i-1], T_H, uls_flag, li_flag);
            print_loop(&L_os[i-1]);
	}
	else if(perm[n][i-1] == li)
	{
            printf("In inverse apply li");
	    li_flag = 1;
            //inverse_apply_li(&L_os[i], &L_os[i-1], T_H);
            inverse_apply_lf(&L_os[i], &L_os[i-1], T_H);
	}
	else if(perm[n][i-1] == uls)
	{
	    uls_flag = 1;
	    inverse_apply_uls(&L_os[i], &L_os[i-1], T_H);
	}
	else
	{
	}
        printf("After applying %d BS, GENERATED LOOP\n", perm[n][i-1]);
        print_loop(&L_os[i-1]);
    } 
    copy_loop(L_ib, &L_os[pos]);
    return ret_val;
}

void permute(int a[], int l, int r, int perm[][MIN_SIZE], int *perm_count)
{
   int c, i;

   if (l == r)
   {
     for(i = 0; i <= r; i++)
     {
         perm[*perm_count][i] = a[i];
     }
     *perm_count = *perm_count + 1;
      //printf("Permi count %d\n", *perm_count);
   }
   else
   {
       for (i = l; i <= r; i++)
       {
          c = a[l]; a[l] = a[i]; a[i] = c;
          permute(a, l+1, r, perm, perm_count);
          c = a[l]; a[l] = a[i]; a[i] = c;
           //backtrack
       } 
   }
}

void copy_T_H_to_sequence(T_H_entry T_H[], int sequence[], int *length_of_sequence)
{
    int i, j;
    j = 0;
    for(i = 0; i < 5; i++)
    {
        if(T_H[i].t == i)
        {
            sequence[j] = i;
            j++;
        }
    }
    *length_of_sequence = j;
}

void generate_permutation_sequence(T_H_entry T_H[], int sequence[], int *length_of_sequence, int perm[][MIN_SIZE], int *number_of_perm)
{
    int i, j;
    copy_T_H_to_sequence(T_H, sequence, length_of_sequence);
    printf("Length of sequence %d\n", *length_of_sequence);
    permute(sequence, 0, *length_of_sequence - 1, perm, number_of_perm);
    printf("Permutation count %d\n", *number_of_perm);
    for(i = 0; i < *number_of_perm; i++)
    {
	printf("Perm ");
	for(j = 0; j < *length_of_sequence; j++)
	{
            printf("%d", perm[i][j]);
	}
	printf("\n");
    }
}

int compute_distance(loop *L_if, loop *L_ib, int k, int var_name_number[][MIN_SIZE], int lb_number[][MIN_SIZE], int ub_number[][MIN_SIZE], int access_exp_number[][MIN_SIZE][MIN_SIZE])
{
    int i, j;	
    int dist = 0;

    if(L_if->nest_count != L_ib->nest_count)
    {
	printf("NEST COUNT ERROR\n");
        dist++;
    }

    if(L_if->line_count != L_ib->line_count)
    {
	printf("LINE COUNT ERROR\n");
        dist++;
    }

    // Assuming equal number of nest counts
    for(i = 0; i < L_if->nest_count; i++)
    {
        /*if(strcmp(L_if->hdl[i].variable_name, L_ib->hdl[i].variable_name))
	{
	    printf("VAR NAME ERROR\n");
	    var_name_number[k][i] = 1;
            dist++;
	}*/
        if(strcmp(L_if->hdl[i].lower_bound, L_ib->hdl[i].lower_bound))
	{
	    printf("LB ERROR\n");
	    lb_number[k][i] = 1;
            dist++;
	}
        if(strcmp(L_if->hdl[i].upper_bound, L_ib->hdl[i].upper_bound))
	{
	    printf("UB ERROR\n");
	    ub_number[k][i] = 1;
            dist++;
	}
    }
    
    for(i = 0; i < L_if->line_count; i++)
    {
        if(L_if->access_count_per_line[i] != L_ib->access_count_per_line[i])
	{
	    printf("ACCESS COUNT ERROR\n");
            dist++;
	}
	for(j = 0; j < L_if->access_count_per_line[i]; j++)
	{
	    if(strcmp(L_if->access_exp[i][j], L_ib->access_exp[i][j]))
	    {
	        printf("ACCESS EXP ERROR %s %s\n", L_if->access_exp[i][j], L_ib->access_exp[i][j]);
		access_exp_number[k][i][j] = 1;
	        dist++;
	    }
	}
    }
    return dist;
}

//void print_details(loop *L_t, int min_dist_val, int length_of_sequence, int perm[][MIN_SIZE], int perm_index[], int min_index[], int j, int var_name_number[][MIN_SIZE], int lb_number[][MIN_SIZE], int ub_number[][MIN_SIZE], int access_exp_number[][MIN_SIZE][MIN_SIZE])
void print_details(loop *L_t, int min_dist_val, int length_of_sequence, int perm[][MIN_SIZE], int min_index[], int j, int var_name_number[][MIN_SIZE], int lb_number[][MIN_SIZE], int ub_number[][MIN_SIZE], int access_exp_number[][MIN_SIZE][MIN_SIZE])
{
    int k, i, l;

    printf("The count of sequences generating minimum distances %d is %d\n", min_dist_val, j);
    for(l = 0; l < j; l++)
    {
	printf("Sequence %d\n", l+1);
        for(i = 0; i < length_of_sequence; i++)
	{
	    //if(perm[perm_index[min_index[l]]][i] == 0)
	    if(perm[min_index[l]][i] == 0)
	    {
	        printf("ld ");
	    }
	    //else if(perm[perm_index[min_index[l]]][i] == 1)
	    else if(perm[min_index[l]][i] == 1)
	    {
	        printf("lf ");
	    }
	    //else if(perm[perm_index[min_index[l]]][i] == 2)
	    else if(perm[min_index[l]][i] == 2)
            {
	        printf("lr ");
	    }
	    //else if(perm[perm_index[min_index[l]]][i] == 3)
	    else if(perm[min_index[l]][i] == 3)
	    {
	        printf("li ");
	    }  
	    //else if(perm[perm_index[min_index[l]]][i] == 4)
	    else if(perm[min_index[l]][i] == 4)
	    {
	        printf("uls ");
	    }  
	    else
	    {
	    }
	}
	printf("\n");
    
        for(k = 0; k < L_t->nest_count; k++)
        {
           if(var_name_number[min_index[l]][k] != 0)
           {
               printf("%dth nest, variable name error\n", k);
           } 
           if(lb_number[min_index[l]][k] != 0)
           {
               printf("%dth nest, lower bound error\n", k);
           } 
           if(ub_number[min_index[l]][k] != 0)
           {
               printf("%dth nest, upper bound error\n", k);
           } 
        }
        for(k = 0; k < L_t->line_count; k++)
        {
            for(i = 0; i < L_t->access_count_per_line[k]; i++)
            {
                if(access_exp_number[min_index[l]][k][i] != 0)
                {
                    printf("%dth line, %dth access expression error\n", k, i);
                }
            }
        }
    }
}

void replace_L_if_by_M_H_M_B(loop *L_if, T_H_entry T_H[], mapping *M_H, mapping *M_B)
{
    int i, j, k, l, m;
    mapping temp_mapping;
    char str[MIN_SIZE], str1[MIN_SIZE];

    for(i = 0; i < M_H->count; i++)
    {
        if(strcmp(M_H->domain[i], M_H->codomain[i]))
	{
	    break;
	}
    }
    if(i == M_H->count)
    {
        return;
    }

    printf("Why ?\n");
    print_mapping(M_H);
    for(i = 0; i < M_H->count; i++)
    {
	printf("M_H->mapping domain %s codomain %s\n", M_H->domain[i], M_H->codomain[i]);
	strcpy(temp_mapping.domain[i], M_H->domain[i]);
	strcpy(temp_mapping.codomain[i], M_H->codomain[i]);
        if(T_H[li].num_of_params == 2)
	{
	    if(T_H[li].param[0] == i)
	    {
	        strcpy(temp_mapping.codomain[i], M_H->codomain[T_H[li].param[1]]);
	    }
            else if(T_H[li].param[1] == i)
	    {
	        strcpy(temp_mapping.codomain[i], M_H->codomain[T_H[li].param[0]]);
	    }
	}
	printf("Temp->mapping domain %s codomain %s\n", temp_mapping.domain[i], temp_mapping.codomain[i]);
    }
    for(j = 0; j < L_if->nest_count; j++)
    { 
        for(i = 0; i < M_H->count; i++)
        {
	    printf("Entered for replacement M_H->domain[i] %s strlen %d M_H->codomain[i] %s strlen %d\n", M_H->domain[i], strlen(M_H->domain[i]), M_H->codomain[i], strlen(M_H->codomain[i]));

	    printf("Original vn %s lb %s ub %s\n", L_if->hdl[j].variable_name, L_if->hdl[j].lower_bound, L_if->hdl[j].upper_bound);
            printf("strstr(L_if->hdl[j].variable_name, M_H->domain[i]) %s\n strstr(L_if->hdl[j].lower_bound, M_H->domain[i]) %s\n strstr(L_if->hdl[j].upper_bound, M_H->domain[i]) %s\n", strstr(L_if->hdl[j].variable_name, M_H->domain[i]), strstr(L_if->hdl[j].lower_bound, M_H->domain[i]), strstr(L_if->hdl[j].upper_bound, M_H->domain[i]));
	    replace_str_me(L_if->hdl[j].variable_name, M_H->domain[i], M_H->codomain[i]);
	    if(strstr(L_if->hdl[j].lower_bound, "max"))
	    {
		m = 0;
	        for(l = 4; l < strlen(L_if->hdl[j].lower_bound) - 1; l++, m++)
		{
		    str[m] = L_if->hdl[j].lower_bound[l];
		}
		str[m] = '\0';
	        replace_str_me(str, temp_mapping.domain[i], temp_mapping.codomain[i]);
		strcpy(L_if->hdl[j].lower_bound, "max(");
		for(l = 0; l < strlen(str); l++)
		{
		    L_if->hdl[j].lower_bound[4 + l] = str[l]; 
		}
		L_if->hdl[j].lower_bound[4 + l] = ')';
		L_if->hdl[j].lower_bound[5 + l] = '\0';
	    }
	    else
	    {
	        replace_str_me(L_if->hdl[j].lower_bound, temp_mapping.domain[i], temp_mapping.codomain[i]);
	    }
	    if(strstr(L_if->hdl[j].upper_bound, "min"))
	    {
	        m = 0;
	        for(l = 4; l < strlen(L_if->hdl[j].upper_bound) - 1; l++, m++)
		{
		    str[m] = L_if->hdl[j].upper_bound[l];
		}
		str[m] = '\0';
	        replace_str_me(str, temp_mapping.domain[i], temp_mapping.codomain[i]);
		strcpy(L_if->hdl[j].upper_bound, "min(");
		for(l = 0; l < strlen(str); l++)
		{
		    L_if->hdl[j].upper_bound[4 + l] = str[l]; 
		}
                L_if->hdl[j].upper_bound[4 + l] = ')';
		L_if->hdl[j].upper_bound[5 + l] = '\0';
            }
	    else
	    {
	        replace_str_me(L_if->hdl[j].upper_bound, M_H->domain[i], temp_mapping.codomain[i]);
	    }
	    printf("Replaced vn %s lb %s ub %s\n", L_if->hdl[j].variable_name, L_if->hdl[j].lower_bound, L_if->hdl[j].upper_bound);
	 }
    }
    for(j = 0; j < L_if->line_count; j++)
    {
        for(k = 0; k < L_if->access_count_per_line[j]; k++)
	{
            for(i = 0; i < M_H->count; i++)
	    {
	        printf("Original exp %s\n", L_if->access_exp[j][k]);
	        replace_str_me(L_if->access_exp[j][k], temp_mapping.domain[i], temp_mapping.codomain[i]);
	        printf("Replaced exp %s\n", L_if->access_exp[j][k]);
	    }
	}
    }
    print_loop(L_if);
}



void find_error_sequence_and_pos(loop *L_s, loop *L_t, T_H_entry T_H[], mapping *M_H, mapping *M_B, int perm[][MIN_SIZE], int number_of_perm, int length_of_sequence)
{
    int i, j, k;
    int ret_val;
    loop L_if, L_ib;
    int distance[MIN_SIZE], perm_index[MIN_SIZE];
    int min_dist_val, min_index[MIN_SIZE];
    int lb_number[MIN_SIZE][MIN_SIZE], ub_number[MIN_SIZE][MIN_SIZE], var_name_number[MIN_SIZE][MIN_SIZE], access_exp_number[MIN_SIZE][MIN_SIZE][MIN_SIZE];
    
    for(i = 0; i < number_of_perm; i++)
    {
	printf("%dth permutation\n", i);
	for(j = 0; j < length_of_sequence; j++)
	{
	    printf("%d", perm[i][j]);
	}
	printf("\n");
	printf("Here\n");
	for(j = 0; j < L_s->nest_count; j++)
	{
	    var_name_number[i][j] = lb_number[i][j] = ub_number[i][j] = 0;
	}
	for(j = 0; j < L_s->line_count; j++)
	{
	    for(k = 0; k < L_s->access_count_per_line[j]; k++)
	    {
	        access_exp_number[i][j][k] = 0;
	    }
	}
        ret_val = construct_forward_sequence(L_s, L_t, &L_if, T_H, perm, i, length_of_sequence);
	print_mapping(M_H);
	replace_L_if_by_M_H_M_B(&L_if, T_H, M_H, M_B);
	print_loop(&L_if);
	distance[i] = compute_distance(&L_if, L_t, i, var_name_number, lb_number, ub_number, access_exp_number);
	//perm_index[i] = i;
	printf("DIST %d\n", distance[i]);
    }
    // Finding minimum distance 
    min_dist_val = distance[0];
    for(i = 0; i < number_of_perm; i++)
    {
        if(distance[i] < min_dist_val)
	{
	    min_dist_val = distance[i];
	}
    }
    // Saving which permutation and the corresponding position is the minimum 
    j = 0; 
    for(i = 0; i < number_of_perm; i++)
    {
        if(distance[i] == min_dist_val)
	{
            min_index[j] = i;
	    j++; 
	}
    }
    //print_details(L_t, min_dist_val, length_of_sequence, perm, perm_index, min_index, j, var_name_number, lb_number, ub_number, access_exp_number);
    print_details(L_t, min_dist_val, length_of_sequence, perm, min_index, j, var_name_number, lb_number, ub_number, access_exp_number);
}

int main()
{
    loop L_s, L_t;
    mapping M_H, M_B;
    T_H_entry T_H[5];
    long offset_to_L_es, offset_to_L_et;
    int  i, j, sequence[5], length_of_sequence;
    int number_of_loops_src, number_of_statements_per_loop_src[MIN_SIZE], number_of_loops_tgt, number_of_statements_per_loop_tgt[MIN_SIZE], pos;
    transformation ld_lf_flag;
    int perm[MIN_SIZE][MIN_SIZE], number_of_perm = 0 ;
    double time_taken;
    // measure time taken by this program
    clock_t t;

    t = clock();
    printf("Please keep the loops in error in a file with the name \"loop\"\n");
    printf("In the \"loop\" file, first enter the source loop, insert a line with ----------, then enter the transformed loop and at last insert a line with ----------\n");
    find_loop_pair_in_error(&L_s, &L_t, &number_of_loops_src, number_of_statements_per_loop_src, &number_of_loops_tgt, number_of_statements_per_loop_tgt, &ld_lf_flag);
    readHeaderDetail(&L_s);
    readHeaderDetail(&L_t);
    readBodyDetail(&L_s);
    readBodyDetail(&L_t);
    print_loop(&L_s);
    print_loop(&L_t);
    construct_M_H(&M_H, &L_s, &L_t);
    print_mapping(&M_H);
    construct_M_B(&M_B, &L_s, &L_t);
    print_mapping(&M_B);
    construct_T_H(&M_H, &M_B, &L_s, &L_t, T_H, number_of_loops_src, number_of_statements_per_loop_src, number_of_loops_tgt, number_of_statements_per_loop_tgt, sequence, ld_lf_flag);
    generate_permutation_sequence(T_H, sequence, &length_of_sequence, perm, &number_of_perm);
    //sequence[0] = ld; 
    //sequence[1] = lr; 
    //length_of_sequence = 2;
    //pos = 2;
    printf("DG\n");
    print_mapping(&M_H);
    find_error_sequence_and_pos(&L_s, &L_t, T_H, &M_H, &M_B, perm, number_of_perm, length_of_sequence);
    t = clock() - t;
    time_taken = ((double)t)/CLOCKS_PER_SEC; // in seconds
    printf("fault localizer took %f seconds to execute \n", time_taken);
}  
