#include "head.h"

/*This function populates the variable table i.e., it stores the information
about each variable in the variable array*/
void populateVariableTable(variable *varr, int *vcount, char **arg, int n)
{
	int i, k, idx;
	char *tok = NULL;

	for(i=1; i<n; i++)						//process each variable
	{
		idx = *vcount;	    
		if(strstr(arg[i], "["))				        //if it is a vector
	        {
        	        k = 0;
			tok = strtok(arg[i], "[]");
			strcpy(arg[0], tok);
        	        while(tok != NULL)
        	        {
				tok = strtok(NULL, "[]");
            	                k++;
        	        }

			varr[idx].t = 1;
			strcpy(varr[idx].n, arg[0]);
			varr[idx].r = k-1;
	        }
	        else							//if it is a scalar
	        {
			varr[idx].t = 0;
			strcpy(varr[idx].n, arg[i]);
			varr[idx].r = 0;
	        }
			
		(*vcount)++;						//increment variable count
	}
}

/*This function checks whether a variable is an input variable, an output variable
or a temporary variable*/
void populateIO(variable *varr, int vcount, char **arg, int n)
{
	int i, j;
	char *tok = NULL;

	for(i=1; i<n; i++)
	{	    
		if(strstr(arg[i], "["))
	        {
			tok = strtok(arg[i], "[]");
			for(j=0; j<vcount; j++)
			{
				if(strcmp(varr[j].n, tok) == 0)
				{
					if(strcmp(arg[0], "input") == 0)		//input array
						varr[j].io = 1;
					else if(strcmp(arg[0], "output") == 0)		//output array
						varr[j].io = 2;
				}
			}
		}
	}
}

//This function populates the domain of each statement
void populateDomain(char **arg, int *lcount, char **lower, char **upper, char *iter, int *flag)
{
    int i = 0, j = 1;
    char *tok = NULL;

    if(strcmp(arg[0], "parallel") == 0)
		j++;

    tok = strtok(arg[j], "=");
    while(tok != NULL)
    {
        arg[i] = (char *)malloc(MAX * sizeof(char));
        strcpy(arg[i], tok);
        tok = strtok(NULL, "=");
        i++;
    }

    if(strstr(arg[1], ":"))
    {
	i=5;
	while(arg[1][i] != ':')
		i++;
	arg[1][i] = ',';
    }

    if(strstr(arg[j+1], ":"))
    {
	i=5;
	while(arg[j+1][i] != ':')
		i++;
	arg[j+1][i] = ',';
    }

    lower[*lcount] = (char *)malloc(2*MIN * sizeof(char));
    upper[*lcount] = (char *)malloc(2*MIN * sizeof(char));
    strcpy(lower[*lcount], arg[1]);
    strcpy(upper[*lcount], arg[j+1]);
    iter[*lcount] = arg[0][0];

    if(j == 1)
	flag[*lcount] = 0;
    else
	flag[*lcount] = 1;

    (*lcount)++;
}

//This function deals with the maintenance of conditional statements i.e., if-else statements
void handleIfElse(char **ifcond, int *ifcount, char **elsecond, int *elsecount, char **arg)
{
	int i, j;
	char *tok = NULL;
        printf("IFELSE\n");
	if(strcmp(arg[0], "if") == 0)
	{
		i = 0;
		ifcond[*ifcount] = (char *)malloc(MAX * sizeof(char));
		tok = strtok(arg[1], "{}");
		while(tok != NULL)
		{
			arg[i] = (char *)malloc(MAX * sizeof(char));
			strcpy(arg[i], tok);
			tok = strtok(NULL, "{}");
			i++;
		}
		strcpy(ifcond[(*ifcount)++], arg[0]);
	}
	else
	{
		i = 1;
		j = 1;
		elsecond[*elsecount] = (char *)malloc(MAX * sizeof(char));
		elsecond[*elsecount][0] = ifcond[*ifcount-1][0];
		if(ifcond[*ifcount-1][i++] == '>')
		{
			if(ifcond[*ifcount-1][i++] == '=')
				elsecond[*elsecount][j++] = '<';
			else
			{
				i--;
				elsecond[*elsecount][j++] = '<';
				elsecond[*elsecount][j++] = '=';
			}
		}
		else
		{
			if(ifcond[*ifcount-1][i++] == '=')
				elsecond[*elsecount][j++] = '>';
			else
			{
				i--;
				elsecond[*elsecount][j++] = '>';
				elsecond[*elsecount][j++] = '=';
			}
		}

		while(ifcond[*ifcount-1][i])
			elsecond[*elsecount][j++] = ifcond[*ifcount-1][i++];

		elsecond[*elsecount][j] = '\0';
		(*elsecount)++;
		(*ifcount)--;
	}
}

/*This function populates the statement table i.e., it stores the information
about each statement in the statement array*/
/* sarr = array of statements, varr = array of variables, vcount = number of variables, scount = number of statements, arg = string representation of the statement, lcount = number of surrounding loops, ifcond = array of strings containing if condition, ifcount = number of if statements, elsecond = array of strings containing else condition, elsecount = number of else statements, dom = set of all loop to access mapping kept for all statements, dcount = total number of mappings, lower = lower bound of the surrounding loop, upper = upper bound of the surrounding loop, iter = the name of the loop iterator, flag = indicates whether the loop is a do or a parallel do loop, loopno = the sequential value to indicate loop's position in the code, lc = exp = the string representation of the postfix expression, T = integer variable to represent rhs of the statement, n = number of accesses, lc = already ended loop number (used to assign current loop number) */
void populateStatementTable(variable *varr, int vcount, statement *sarr, int *scount, char **arg, int lcount, char **ifcond, int ifcount, char **elsecond, int elsecount, char **dom, int *dcount, char **lower, char **upper, char *iter, int *flag, int loopno, int lc, int hcount, int *hier)
{
	int i, j, k, l, L = 0, T = 1;
	int ind, idx = *scount;
	char *tok = NULL;
	
	char s[2];
	char ct[2];
	char it[2];
	char con[MAX];
	
	char **str = (char **)malloc(MAX * sizeof(char *));
	char **exp = (char **)malloc(MAX * sizeof(char *));

	arg[0] = strtok(arg[0], ":");
	strcpy(s, arg[0]);

        /* ind is type of statement which can assume values from the set { wait, signal or assignment } */
	ind = decideType(sarr, idx, arg, con, &L, &T);
//printf("L\n");
/*finding L field of statement in case of signal & assignment statement BEGIN*/
	if(ind != 2)
		k = populateL(sarr, idx, varr, vcount, arg, str, exp, L);
/*finding L field of statement in case of signal & assignment statement END*/
//printf("T\n");
/*finding T field of statement in case of wait & assignment statement BEGIN*/
	if(ind != 1)
		k = populateT(sarr, idx, varr, vcount, arg, str, exp, T, &j);
/*finding T field of statement in case of wait & assignment statement END*/
//printf("S\n");
	strcpy(sarr[idx].v.s, "NULL");				//initially the DG vertex is NULL
	sarr[idx].d = lcount;
        for(i = 0; i < MAXNEST; i++)
        {
            sarr[idx].lhier[i] = -1;
        }
        for(l = 1; l <= hcount; l++)
        {
            sarr[idx].lhier[l - 1] = hier[l - 1];
        }
	sarr[idx].lis = loopno - lc;
        // printf("Statement %d loopno %d\n", idx, sarr[idx].lis);
	memset(sarr[idx].D, '\0', sizeof(sarr[idx].D));
	sarr[idx].o = (int *)malloc((lcount) * sizeof(int));
        // printf("Initial sarr[idx].D %s", sarr[idx].D);

	for(i=0; i<lcount; i++)
        {
		sarr[idx].o[i] = flag[i];
        }

/*Populating domain of statement BEGIN*/
        if(sarr[idx].lis != 0)
        {
	    if(sarr[idx].d == 0)
	    {
	    	for(i=0; i<sarr[idx].L -> r; i++)
	    	{
	    		strcat(sarr[idx].D, "100 <= ");
	    		strcat(sarr[idx].D, sarr[idx].L -> P -> P[i] -> n);
	    		strcat(sarr[idx].D, " <= 100");
	    		if(i != (sarr[idx].L -> r)-1)
	    			strcat(sarr[idx].D, " and ");
	    	}
	    }
	    else
	    {
	    	for(i=0; i<lcount; i++)
	    	{
	    		strcat(sarr[idx].D, lower[i]);
	    		if(1 < 100)
	    			strcat(sarr[idx].D, " <= ");
	    		else
	    			strcat(sarr[idx].D, " >= ");

	    		sprintf(it, "%c", iter[i]);
	    		strcat(sarr[idx].D, it);
	    		if(1 < 100)
	    			strcat(sarr[idx].D, " <= ");
	    		else
	    			strcat(sarr[idx].D, " >= ");

	    		strcat(sarr[idx].D, upper[i]);
	    		if(i != lcount-1)
	    			strcat(sarr[idx].D, " and ");
	    	}

	    	if(ind == 1 || ind == 2)
	    	{
                            if(strcmp(con, "") != 0)
                            {
	    		    strcat(sarr[idx].D, " and ");
	    		    strcat(sarr[idx].D, con);
                            }
	    	}
	    }
        }

/* The domain is modified in presence of if and else statements */
	for(i=0; i<ifcount; i++)
	{
		strcat(sarr[idx].D, " and ");
		strcat(sarr[idx].D, ifcond[i]);
	}

	for(i=0; i<elsecount; i++)
	{
		strcat(sarr[idx].D, " and ");
		strcat(sarr[idx].D, elsecond[i]);
	}
        // printf("Final sarr[idx].D %s", sarr[idx].D);

/* Creating surrounding loop to statement range mapping */
	if(ind != 2)
		isccMapping(varr, vcount, sarr, idx, ind, s, iter, lcount, str, dom, dcount, "L", "S", 1);
// printf("LD\n");
	if(ind != 1 && !(sarr[idx].T -> n[0] >= '0' && sarr[idx].T -> n[0] <= '9'))
	{
		k = 1;
		for(i=0; i<j; i++)
		{
			if(i == 0 || !isOperand(arg[i][0]))
			{
				sprintf(ct, "%d", k);
				isccMapping(varr, vcount, sarr, idx, ind, s, iter, lcount, arg, dom, dcount, ct, "W", i+1);
				k++;
			}
		}
	}
        // printf("Final Final sarr[idx].D %s", sarr[idx].D);
/*Populating domain of statement END*/
//printf("D\n");
	(*scount)++;								//increment statement count
	free(str);
	free(exp);
}
    //printf("address1 %u\n", a);
