#include "head.h"

Stack *createStack(unsigned capacity)
{
    Stack *stack = (Stack *)malloc(sizeof(Stack));
    stack -> top = -1;
    stack -> capacity = capacity;
    stack -> array = (char *)malloc(capacity * sizeof(char));
    return stack;
}

int isEmpty(Stack *stack)
{
    return (stack -> top == -1);
}

void push(Stack *stack, char op)
{
    stack -> array[++(stack -> top)] = op;
}

char pop(Stack *stack)
{
    if(isEmpty(stack))
        return '$';
    return stack -> array[(stack -> top)--];
}

char peek(Stack *stack)
{
    if(isEmpty(stack))
        return '$';
	return stack -> array[stack -> top];
}

sNode *newNode(expTreeNode *node)
{
	sNode *newnode = (sNode *)malloc(sizeof(sNode));
	newnode -> node = node;
	newnode -> next = NULL;
	return newnode;
}

int issEmpty(sNode *root)
{
    return !root;
}

void spush(sNode **root, expTreeNode *node)
{
    sNode *newnode = newNode(node);
    newnode -> next = *root;
    *root = newnode;
}

expTreeNode *spop(sNode **root)
{
    if(issEmpty(*root))
        return NULL;

    sNode *top = *root;
    expTreeNode *popped = top -> node;
    *root = top -> next;
    // free(top);
    return popped;
}

int isOperand(char ch)
{
    return ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || (ch >= '0' && ch <= '9'));
}

/* Finding precedence of the operator */
int prec(char ch)
{
    switch(ch)
    {
        case '+':
        case '-':
            return 1;
        case '*':
        case '/':
		case '%':
            return 2;
        case '^':
            return 3;
        case '#':
            return 4;
    }
    return -1;
}

//This function converts an infix expression to a postfix expression
void infixToPostfix(char *str, char* exp)
{
    Stack* stack = createStack(strlen(str));
    if(!stack)
        return;

    int i,j;
    for(i=0, j=-1; str[i]; i++)
    {
        if(isOperand(str[i]))
            exp[++j] = str[i];

        else if(str[i] == '(')
            push(stack, str[i]);

        else if(str[i] == ')')
        {
            exp[++j] = '$';
			while(!isEmpty(stack) && peek(stack) != '(')
                    exp[++j] = pop(stack);

            if(!isEmpty(stack) && peek(stack) != '(')
                return;
            else
                pop(stack);
        }
        else
        {
            if(str[i-1] != ')' && str[i-1] != '(')
				exp[++j] = '$';

			while(!isEmpty(stack) && prec(str[i]) <= prec(peek(stack)))
                exp[++j] = pop(stack);
            push(stack, str[i]);
        }
    }
    if(str[i-1] != ')')
				exp[++j] = '$';

	while(!isEmpty(stack))
        exp[++j] = pop(stack);

    exp[++j] = '\0';
}

//This function finds the inorder traversal of the expression tree
void inorder(expTreeNode *root, expTreeNode **temp, int *idx)
{
	if(root -> r == 0)
		return;

	if(isOperand(root -> n[0]))
	{
		int i;
		for(i=0; i<root -> r; i++)
			root -> P[i] = temp[(*idx)++];
		return;
	}

        if(root -> r == 2)
        {
	    inorder(root -> P[0], temp, idx);
	    inorder(root -> P[1], temp, idx);
        }
        else
        {
	    inorder(root -> P[0], temp, idx);
        }
}

/*This function creates the expression tree corresponding to right hand side of
assignment statement*/
expTreeNode *constructExpressionTree(char **exp, variable *varr, int vcount, int count)
{
	sNode* stack = NULL;
	expTreeNode **temp = (expTreeNode **)malloc(count * sizeof(expTreeNode *));
	int idx, i, j, k, l;
        char s[MAX];

	for(idx=0; idx<count; idx++)
	{
		for(i=0; exp[idx][i]; i++)
    	        {
    	                temp[idx] = (expTreeNode *)malloc(sizeof(expTreeNode));
			l = i;
			j = 0;

			if(isOperand(exp[idx][l]))
			{
				while(exp[idx][i] != '$')
					s[j++] = exp[idx][i++];
				s[j] = '\0';

				if(exp[idx][l] >= '0' && exp[idx][l] <= '9')
				{
					temp[idx] -> t = 3;
					temp[idx] -> r = 0;
				}
				else
				{
					for(k=0; k<vcount; k++)
					{
						if(strcmp(s, varr[k].n) == 0)
						{
							temp[idx] -> t = varr[k].t;
							temp[idx] -> r = varr[k].r;
						}
					}
				}
			}
			else
			{
				s[j++] = exp[idx][l];
				s[j] = '\0';
				temp[idx] -> t = 2;
                                if(exp[idx][l] == '#')
				    temp[idx] -> r = 1;
                                else
				    temp[idx] -> r = 2;
			}

			strcpy(temp[idx] -> n, s);

			temp[idx] -> P = (expTreeNode **)malloc((temp[idx] -> r) * sizeof(expTreeNode *));
		
			if(isOperand(exp[idx][l]))
				spush(&stack, temp[idx]);
			else
			{
				for(k=(temp[idx] -> r)-1; k>=0; k--)
					temp[idx] -> P[k] = spop(&stack);
				spush(&stack, temp[idx]);
			}
		}
		temp[idx] = spop(&stack);
	}

	k = 1;
	expTreeNode *root = temp[0];
	inorder(root, temp, &k);
	return root;
}

/*This function decides the type of statement whether it is a wait statement,
a signal statement or an assignment statement*/
int decideType(statement *sarr, int idx, char **arg, char *con, int *L, int *T)
{
	int j, ind = 0, cond_flag = 0;
	char *tok = NULL;

        strcpy(con, "");
        if(strstr(arg[1], "if"))
        {
            cond_flag = 1;
        }

	if(strstr(arg[1], "signal"))				//signal statement
	{
		j = 0;
       	        tok = strtok(arg[1], "()");
                while(tok != NULL)
                {
                    arg[j] = (char *)malloc(MAX * sizeof(char));
                    strcpy(arg[j], tok);
                    tok = strtok(NULL, "()");
                    j++;
                }
		ind = 1;
		sarr[idx].t = 2;
                if(cond_flag == 1)
                {
                    *L = 3;
		    strcpy(con, arg[1]);
                }
                else
                {
                    *L = 1;
                }
	}
	else if(strstr(arg[1], "wait"))				//wait statement
	{		
		j = 0;
       	        tok = strtok(arg[1], "()");
                while(tok != NULL)
                {
                    arg[j] = (char *)malloc(MAX * sizeof(char));
                    strcpy(arg[j], tok);
                    tok = strtok(NULL, "()");
                    j++;
                }
		ind = 2;
		sarr[idx].t = 1;
                if(cond_flag == 1)
                {
                    *T = 3;
		    strcpy(con, arg[1]);
                }
                else
                {
                    *T = 1;
                }
	}
	else										//assignment statement
	{
		j = 0;
	        tok = strtok(arg[1], "=");
    	        while(tok != NULL)
    	        {
    		    arg[j] = (char *)malloc(MAX * sizeof(char));
   	    	    strcpy(arg[j], tok);
     		    tok = strtok(NULL, "=");
        	    j++;
    	        }
		sarr[idx].t = 0;
	}
	return ind;
}

//This function populates the left hand side of an assignment statement
/* idx = index of the current statement, arg = string representation of the statement, str = not used here, exp = the postfix expression of the lhs to create expression tree, L = indicator to state whether it is an assignment statement or signal statement */
int populateL(statement *sarr, int idx, variable *varr, int vcount, char **arg, char **str, char **exp, int L)
{
	int i, j = 0, k = 0;
	char *tok = NULL;
	scaVecRef *ref = (scaVecRef *)malloc(sizeof(scaVecRef));

	tok = strtok(arg[L], "[]");
	while(tok != NULL)
	{
	    str[j] = (char *)malloc(MIN * sizeof(char));
	    strcpy(str[j], tok);
	    tok = strtok(NULL, "[]");
	    j++;
	}

	for(i=0; i<vcount; i++)
	{
		if(strcmp(str[0], varr[i].n) == 0)
		{
			ref -> t = varr[i].t;
			ref -> r = varr[i].r;
		}
	}

	strcpy(ref -> n, str[0]);

	if(ref -> t == 0)
		ref -> P = NULL;
	else
	{
		ref -> P = (expTreeNode *)malloc(sizeof(expTreeNode));
		strcpy(ref -> P -> n, ref -> n);
		ref -> P -> t = ref -> t;
		ref -> P -> r = ref -> r;
		ref -> P -> P = (expTreeNode **)malloc(ref -> r * sizeof(expTreeNode *));

		exp[k] = (char *)malloc(MAX * sizeof(char));
		for(i=0; i<ref -> r; i++)
		{
			infixToPostfix(str[i+1], exp[k]);
			ref -> P -> P[i] = constructExpressionTree(exp, varr, vcount, 1);
		}
	}

        /* Finally populates the L field of a statement */
	sarr[idx].L = ref;
	return k;
}

//This function populates the right hand side of an assignment statement
/* sarr = array of statements, idx = index of the current statement, varr = array of variables, vcount = number of variables, arg = string representation of the statement, exp = the string representation of the postfix expression, T = integer variable to represent rhs of the statement, n = number of accesses */
int populateT(statement *sarr, int idx, variable *varr, int vcount, char **arg, char **str, char **exp, int T, int *n)
{
	int i, j = 0, k = 0;
	char *tok = NULL;
	char temp[MAX];

	if(strstr(arg[T], "["))
	{
	   	tok = strtok(arg[T], "[]");
	        while(tok != NULL)
	        {
	            arg[j] = (char *)malloc(MAX * sizeof(char));
	            strcpy(arg[j], tok);
	            tok = strtok(NULL, "[]");
	            j++;
	        }

		strcpy(temp, arg[0]);
		for(i=1; i<j; i++)
			if(!isOperand(arg[i][0]))
				strcat(temp, arg[i]);

		exp[k] = (char *)malloc(MAX * sizeof(char));
		infixToPostfix(temp, exp[k]);
                //if(exp[k][strlen(exp[k]) - 1]  == '$')
                //    strcat(exp[k], "#");
		for(i=1; i<j; i++)
		{
			if(isOperand(arg[i][0]))
			{
				exp[++k] = (char *)malloc(MAX * sizeof(char));
				infixToPostfix(arg[i], exp[k]);
			}
		}
                
	}
	else
	{
		j = 2;
		exp[k] = (char *)malloc(MAX * sizeof(char));
		infixToPostfix(arg[T], exp[k]);
	}

	sarr[idx].T = constructExpressionTree(exp, varr, vcount, k+1);
	*n = j;
	return k;
}
