#include<string.h>
#include<stdlib.h>
#include<stdio.h>
#include<math.h>
#include<time.h>
#include<sys/time.h>
#define MAX 1000
#define MIN 100
#define MAXNEST 3

double startcputime, endcputime;
/*Expression Tree Node Structure*/
typedef struct expTreeNode
{
    int t;                      //type, 0 - scalar, 1 - vector, 2 - operation, 3 - function call
    char n[MIN];               	//name of the operation
    int r;                      //arity of the operation, 0 - const, 1 - unary, 2 - binary, dim - vector, x - scalar
    struct expTreeNode **P;     //r - dim ptr array to child nodes
}expTreeNode;

/*Variable Strcture*/
typedef struct variable
{
    int t;                      //type, 0 - scalar, 1 - vector
    char n[MIN];                //name
    int r;                   	//arity, dim - vector, x - scalar
    int io;			//0 - default, 1 - input, 2 - output

    int  uninit_varr;
    char yaml_ref[4*MAX];
    char yaml_name[4*MAX];
    char yaml_type[4*MAX];
    char alt_yaml_ref[4*MAX];
}variable;

/*Scalar Vector Reference Structure*/
typedef struct scaVecRef
{
	int t;					//type, 0 - scalar, 1 - vector
	char n[MIN];				//name
	int r;					//arity, dim - vector, x - scalar
	expTreeNode *P;				//r - dim ptr to index expression tree
}scaVecRef;

/*Set Structure*/
typedef struct set
{
	char s[MIN];				//statement no
	char D[23*MAX];				//domain
        char DG_D[23*MAX];                      //domain actually kept in DG 
}set;

/*Access Structure*/
typedef struct access
{
	char s[MIN];				//statement no
	expTreeNode *r;				//reference to set of memory locations
	int t;					//type of access, 0 - read, 1 - write
	char R[3*MAX];				//range
	set *S;					//set of ordered pairs depicting S<alpha>

        char yaml_name[2*MAX];
        int  var_idx;
        int  uninit_flag;                       //some range of the range of the access is uninitialized
        char uninit_R[2*MAX];                          //uninitialized range
        int  WAW_edge;                         // WAW edge (if any) around it 
        int  st_idx;                           // statement index
}access;

/*Conflict Edge Structure*/
typedef struct conf_edge
{
	access a1;				//access 1
	access a2;				//access 2
	int t;					//type of edge, 0 - RAW, 1 - WAR, 2 - WAW
	char M[5*MAX];				//dependence mapping
	int flag;				//flag, 0-excluded, 1-included
	int dir;				//flag, 0-undirected, 1-directed

	char included_M[2*MAX];			// included dependence mapping
}conf_edge;

/*Vertex Structure*/
typedef struct vertex
{
	char s[MIN];				//statement no
	char f[MIN];				//operation
	int r;						//arity
	int d;						//loop depth
	char D[2*MAX];				//domain
}vertex;

/*Dependence Edge Structure*/
typedef struct dep_edge
{
	vertex s;					//source vertex
	vertex t;					//target vertex
	int p;						//position
	char M[2*MAX];				//mapping
}dep_edge;

/*Statement Structure*/
typedef struct statement
{
        int t;				//type, 0 - assignment, 1 - wait, 2 - signal
	scaVecRef *L;         		//ptr to variable
        expTreeNode *T;          	//ptr to parse tree
        vertex v;                       //DG vertex
        int d;                          //nesting depth of loop surrounding s, 0 - if no loop
        char D[2*MAX];                  //domain of s, x - if no loop
	int *o;				//flag, 0 - sequential, 1 - parallel
	int lis;			//loop no
        int lhier[MAXNEST];             //loop hierarchy

        char yaml_name[2*MAX];
        int  yaml_flag;
}statement;

/*Path Structure*/
typedef struct path
{
	access *P;					//path
        int n;                                  // number of accesses in a path 
	char M[2*MAX];				//composition of all the mappings of the edges
}path;

/*Stack 1 for infix to postfix conversion*/
typedef struct Stack
{
    int top;
    char *array;
    unsigned capacity;
}Stack;

/*Stack 2 for Expression tree construction*/
typedef struct sNode
{
	expTreeNode *node;
	struct sNode *next;
}sNode;


typedef struct eStack
{
    int top;
    int array[MAX];
}eStack;


typedef struct etNode
{
    int v1;
    int v2;
    char Rwant[5*MAX];
    char Rlost[5*MAX];
    char Rneed[5*MAX];
    struct etNode *children;
}etNode;


typedef struct etEdge
{
    int n1; // Dependence graph edge index of the source and transformed programs subjected to the equivalence checking procedure
    int n2;
}etEdge;


typedef struct recurEqn
{
    int d3;
    int d4;

    int c1;
    int c2;
}recurEqn;

typedef struct recurEqnSoln
{
    int baseVal;
    int recurEqnIndex;
    int recurEqnSubIndex;

    // im = i0*c1^m + Sm where Sm = m.c2 (if c1 = 1) or (-c2/(1 - c1))*(c1^m - 1)
    // alpha is the coefficient associated with c1^m and beta is the rest of the part in the solution 
    int alpha;
    int beta;
}recurEqnSoln;

typedef struct combinedRecurEqn
{
    int baseVal;
    int coeff[100];
    int coeffCount;

}combinedRecurEqn;

void parseInput(char* file_name, variable *varr, int *vcount, statement *sarr, int *scount, char **dom, int *dcount);

int populateAccessTable(access *aarr, statement *sarr, int scount, char **dom, variable *varr, int vcount);

int construct_cagp(variable *varr, access *aarr, int acount, conf_edge *carr, statement *sarr, char **dom, int dcount);

void printConflictEdgeTable(conf_edge *carr, int ecount);

int populateWaitSignal(access *aws, statement *sarr, access *aarr, int acount);

void construct_ws(access *aarr, int acount, statement *sarr, conf_edge *carr, int *ecount, access *aws, int count, conf_edge *ews[], conf_edge *fsw[], char **dom);

void populateReadAccess(variable *varr, int vcount, statement *sarr, access *aarr, int acount, conf_edge *carr, int ecount, char **dom, int dcount);

void populateWriteAccess(variable *varr, int vcount, statement *sarr, access *aarr, int acount, conf_edge *carr, int ecount);

void findPaths(access *aws, int count, statement *sarr, conf_edge *ews[], conf_edge *fsw[], path *pws, int *pcount, path *cws, int *ccount);

void undirected(access *aarr, int acount, statement *sarr, conf_edge *carr, int *ecount, path *pws, int pcount, path *cws, int ccount, conf_edge *ews[], int count, char **dom);

void checkDeterminacy(conf_edge *carr, int ecount);

void construct_dg(variable *varr, int vcount, statement *sarr, int scount, access *aarr, int acount, conf_edge *carr, int ecount, vertex *dvarr, int *dvcount, dep_edge *dearr, int *decount, char **dom , int dcount);

void printVariableTable(variable *varr, int vcount);

void printStatementTable(statement *sarr, int scount);

void printAccessTable(access *aarr, int acount);

void printPathsCycles(path *pws, int pcount, path *cws, int ccount);

void printVertexTable(vertex *dvarr, int dvcount);

void printDependenceEdgeTable(dep_edge *dearr, int decount);

void print_yaml_file(char *file_name, variable *varr, int vcount, access *aarr, int acount, conf_edge *carr, int ecount, statement *sarr, int scount, char **dom, int dcount);

void generalize(char log_file_name[], vertex *dvarr1, int dvcount1, vertex *dvarr2, int dvcount2, dep_edge *dearr1, int decount1, dep_edge *dearr2, int decount2, variable *varr, int vcount, statement *sarr, int scount, access *aarr, int acount);

void populateVariableTable(variable *varr, int *vcount, char **arg, int n);

void populateIO(variable *varr, int vcount, char **arg, int n);

void populateDomain(char **arg, int *lcount, char **lower, char **upper, char *iter, int *flag);

void handleIfElse(char **ifcond, int *ifcount, char **elsecond, int *elsecount, char **arg);

void populateStatementTable(variable *varr, int vcount, statement *sarr, int *scount, char **arg, int lcount, char **ifcond, int ifcount, char **elsecond, int elsecount, char **dom, int *dcount, char **lower, char **upper, char *iter, int *flag, int loopno, int lc, int hcount, int *hier);

int decideType(statement *sarr, int idx, char **arg, char *con, int *L, int *T);

int populateL(statement *sarr, int idx, variable *varr, int vcount, char **arg, char **str, char **exp, int L);

int populateT(statement *sarr, int idx, variable *varr, int vcount, char **arg, char **str, char **exp, int T, int *n);

void isccMapping(variable *varr, int vcount, statement *sarr, int idx, int ind, char *s, char *iter, int lcount, char **str, char **dom, int *dcount, char *s1, char *s2, int j);

int isOperand(char ch);

void fileHandling(char *str);

void isccInterSub(char *temp, char *str1, char *str2, char *op);

void isccComposition(access *aarr, int acount, access a1, access a2, char *temp, char **dom);

void isccDomain(char *temp, char *str);

void extractStr(char *D, char *delim1, char *delim2, char *str);

int populateIV(char **iv, char *tem);

void createEdge(access *aarr, int acount, conf_edge *carr, access a1, access a2, char **dom, char *temp, int idx);

void modify(char *temp, char *mapping, char *str);

void parse(char *str, char *tem, int ind);
