#include "head.h"

//This function finds the gcd of two integers
int gcd(int x, int y)
{
    if(y == 0)
        return x;
    else
        return gcd(y, x%y);
}

//This function shifts the cycle by a fixed no of vertices
void cycleShift(access *arr, int d, int n)
{
    int i, j, k;
	access temp;

    for(i=0; i<gcd(d,n); i++)
    {
        temp = arr[i];
        j = i;
        while(1)
        {
            k = j + d;
            if(k >= n)
                k = k - n;

            if(k == i)
                break;

            arr[j] = arr[k];
            j = k;
        }
        arr[j] = temp;
    }
}

//This function creates the mapping of the created subpath of a bigger path
/* Input : p = subpath, idx = length of the subpath, ews = set of wait-signal conflict edges, count = number of edges  */
void createPath(conf_edge *ews[], int count, path *p, int idx)
{
	int k, l = 1, m, n, ind = 0;
	char it[3];
	char temp[3*MAX];

	memset(temp, '\0', sizeof(temp));
	for(k=0; k<idx; k+=2)
	{
		for(m=0; m<count; m++)
		{
			for(n=0; n<count; n++)
				if(ews[m][n].flag == 1 && ews[m][n].dir == 1 && strcmp(ews[m][n].a1.s, p->P[k].s) == 0 && strcmp(ews[m][n].a2.s, p->P[k+1].s) == 0)
					break;
			if(n < count)
				break;
		}

		strcat(temp, "M");
		sprintf(it, "%d", l);
		strcat(temp, it);
		strcat(temp, " := ");
		strcat(temp, ews[m][n].M);
		strcat(temp, ";");
		l++;
	}

	strcat(temp, "M := M1");
	for(k=2; k<l; k++)
	{
		strcat(temp, "( M");
		sprintf(it, "%d", k);
		strcat(temp, it);
		ind++;
	}

	for(k=0; k<ind; k++)
		strcat(temp, " )");
	strcat(temp, ";M;");

	fileHandling(temp);
	strcpy(p->M, temp);
}

/* Extracts coefficients of the mapping */
/* Input : str = a mapping e.g. {ai + b | 1 <= i <= N}, Output : *c = a, *d = b*/
void opt(char *str, int *c, int *d)
{
	int i = 0;
	char it[3];

	*c = 1;
	if(str[10] == 'i')
			*d = 0;
	else
	{
		while(str[i+10] != ' ')
		{
			it[i] = str[i+10];
			i++;
		}
		it[i] = '\0';
		*d = atoi(it);
	}
}

/*This function calculates the value of k where k is the no of times a certain
portion of walk is repeated in order to make this a supporting walk
and this k should be an integer*/
int calculate_k(int a1, int b1, int a2, int b2, int a3, int b3, int c, int d)
{
	int t;
	double k;

	if(a3 == 1)
	{
		k = ((double)d - (double)a2*(double)b1 - (double)b2)/((double)a2*(double)b3);
		t = (int)k;

		if(k == t)
			return 1;
		else
			return 0;
	}
	else
	{
		k = log((double)c/((double)a1*(double)a2))/log((double)a3);
		t = (int)k;

		if(k == t && d == (a2*pow(a3,t)*b1 + a2*b3*((pow(a3,t)-1)/(a3-1)) + b2))
			return 1;
		else
			return 0;
	}
}

//This function finds supporting path if present
void findSupportingPath(access a1, access a2, statement *sarr, path *pws, int pcount, path *tpws, int *tpcount, char *domain, char *mapping)
{
	int i, j, k;

	char dom[MAX];
	char str[MAX], str1[3*MAX];
	char temp[3*MAX];
	char ar1[MAX];
	char ar2[MAX];
  
        printf("Begin findSupportingPath\n");
	for(i=0; i<pcount; i++)
	{
		j = 0;
		while(pws[i].P[j].s[0] == 'S')
			j++; /*Finding the last access in a path*/

		if(a1.st_idx > pws[i].P[0].st_idx && a2.st_idx < pws[i].P[j-1].st_idx)
		{ 
                        /*Comparison of statement number to find the control-flow*/
                        /*The function parse is written in the file variable_table.c from where the association rule of a mapping is returned*/
			/*parse(str, pws[i].M, 0);
			strcpy(ar1, str);

			parse(str, mapping, 0);
			strcpy(ar2, str);

			if(strcmp(ar1, ar2) == 0)
			{
                                //Checking whether the mappings are equal for the path and the assumed mapping
				isccDomain(temp, pws[i].M);
				fileHandling(temp);
				strcpy(dom, temp);

				isccInterSub(temp, dom, domain, "-");
				fileHandling(temp);
				strcpy(domain, temp);
			}
			else
				tpws[(*tpcount)++] = pws[i];*/ /*PI' is updated*/

                        strcpy(str, "M1 := ");
                        strcat(str, mapping);
                        strcat(str, ";");
                        sprintf(str1, "M2 := %s; M3 := M1 - M2; S := dom(M3); S;", pws[i].M);
                        strcat(str, str1);
                        fileHandling(str);
                        strcpy(domain, str);
			if(strcmp(domain, "{  }") == 0)
				break;
		}
	}
        printf("End findSupportingPath\n");
}

//This function finds supporting walk if present
/* This function is the same as FindSupportingWalk in the paper*/
void findSupportingPathWithCycle(access a1, access a2, statement *sarr, path *tpws, int tpcount, path *cws, int ccount, conf_edge *ews[], int count, char *domain, char *mapping)
{
	int i, j, k, l, m, n, val;
	int idx1, idx2;
	int len1, len2, ind;
	int c1, d1, c2, d2, c3, d3, c4, d4;

	char it[3];
	char dom[3*MAX], temp[3*MAX], str[3*MAX], temp_str[3*MAX];
	path p1, p2, c;

        printf("Begin findSupportingPathWithCycle\n");
	for(i=0; i<tpcount; i++)
	{
		len1 = 0;
		while(tpws[i].P[len1].s[0] == 'S')
			len1++;
		//printf("len1 - %d\n", len1);
		if(a1.st_idx > tpws[i].P[0].st_idx && a2.st_idx < tpws[i].P[len1-1].st_idx)
		{
			for(j=0; j<ccount; j++)
			{
				len2 = 0;
				while(cws[j].P[len2].s[0] == 'S')
					len2++;
				//printf("len2 - %d\n", len2);
				for(idx1=0; idx1<len1; idx1++)
				{
                                        /* Finding whether the path tpws[i] and the cycle cws[j] are overlapping or not */
					for(idx2=0; idx2<len2; idx2++)
						if(strcmp(tpws[i].P[idx1].s, cws[j].P[idx2].s) == 0)
							break;
					if(idx2 < len2)
							break;
				}

				if(idx1 < len1 && idx2 < len2)
				{//printf("idx1 - %d idx2 - %d\n", idx1, idx2);
                                        /*There is an overlap between the path and the cycle; hence two subpaths p1 and p2 are created*/
					if(idx1 != 0)
					{
						p1.P = (access *)malloc(idx1 * sizeof(access));
						for(k=0; k<idx1; k++)
							p1.P[k] = tpws[i].P[k];

						createPath(ews, count, &p1, idx1);
					}

                                        /*len1 is the length of the path tpws[i], idx1 is the length of p1, len1 - idx1 is the length of p2*/
					if(idx1 != len1)
					{
						p2.P = (access *)malloc((len1-idx1) * sizeof(access));
						for(k=idx1; k<len1; k++)
							p2.P[k-idx1] = tpws[i].P[k];
//printf("%s %s\n", p2.P[0].s, p2.P[1].s);
                                                /* Create path assigns mapping of the created subpath*/
						createPath(ews, count, &p2, len1-idx1);
					}

					c.P = (access *)malloc((len2-1) * sizeof(access));
					for(m=0; m<len2-1; m++)
						c.P[m] = cws[j].P[m];

					cycleShift(c.P, idx2, m);							//Calling cycle shift function
/*createPath is called to assign mapping of the cyclically shifted cycle*/
					createPath(ews, count, &c, len2-1);

					if(idx1 == 0)
						c1 = d1 = 0;
					else
                                                /* extracts a and b from a mapping of the form { ai + b }; here c1 = a and d1 = b*/
						opt(p1.M, &c1, &d1);

					if(idx1 == len1)
						c2 = d2 = 0;
					else
						opt(p2.M, &c2, &d2);
					
					opt(c.M, &c3, &d3);
					opt(mapping, &c4, &d4);
                                        val = calculate_k(c1, d1, c2, d2, c3, d3, c4, d4); //calling calculate_k function

                                        if(val > 0)    // val is k' in the paper
                                        {
                                            strcpy(str, "M1 := ");
                                            strcat(str, p1.M);
                                            strcat(str, ";");
                                            strcat(str, "M2 := ");
                                            strcat(str, c.M);
                                            sprintf(temp_str, "^%d", val);
                                            strcat(str, temp_str);
                                            strcat(str, ";");
                                            strcat(str, "M3 := ");
                                            strcat(str, p2.M);
                                            strcat(str, "; M4 := M1(M2); M := M4(M3); M;");
				            fileHandling(str);
                                            isccDomain(temp, str);
				            fileHandling(temp);
				            strcpy(dom, temp);

				            isccInterSub(temp, dom, domain, "-");
				            fileHandling(temp);
				            strcpy(domain, temp);
                                            return;
                                        }

				}
			}
		}
	}
        printf("End findSupportingPathWithCycle\n");
}

/*This function finds supporting path or supporting walk if present*/
/* The function is the same as InstallEdge function in the paper */
void supported(access a1, access a2, statement *sarr, path *pws, int pcount, path *cws, int ccount, conf_edge *ews[], int count, char *domain, char *mapping)
{
	int i, j, k;
	int tpcount = 0;
	path *tpws = (path *)malloc(MAX * sizeof(path));

        printf("sdutta : supported for mapping %s\n", mapping);
        /* As per paper, Input of FindSupportingPath : alpha = a1, alpha' = a2, PI = pws, pcount,  Du = domain; Output : Du = domain, PI' = tpws, tpcount */
	findSupportingPath(a1, a2, sarr, pws, pcount, tpws, &tpcount, domain, mapping);
//printf("domain - %s\n", domain);printf("mapping - %s\n", mapping);
	if(strcmp(domain, "{  }") != 0)
	{//printf("tpcount - %d\n", tpcount);
		if(tpcount != 0)
/*As per paper, Input of FindSupportingPathWithCycle : alpha = a1, alpha' = a2, PI = pws, pcount,  C = cws, ccount, Du = domain; Output : Du = domain*/
			findSupportingPathWithCycle(a1, a2, sarr, tpws, tpcount, cws, ccount, ews, count, domain, mapping);
	}
        printf("End supported domain %s\n", domain);
}

//This function assigns direction to undirected edges
/*aarr = access array, acount = access count, carr = conflict edge array, ecount = conflict edge count, pws = set of paths, pcount = number of paths, cws = set of cycles, ccount = number of cycles, ews = conflict edges in the wait-signal graph, count = number of wait-signal accesses, dom = array of strings containing domains of all the statements*/
void undirected(access *aarr, int acount, statement *sarr, conf_edge *carr, int *ecount, path *pws, int pcount, path *cws, int ccount, conf_edge *ews[], int count, char **dom)
{
	int i, j;
	int cnt = *ecount;

	conf_edge tem;
	char str[MAX];
	char temp[3*MAX];
	char domain[MAX];
	char mapping[MAX];

	for(i=0; i<cnt; i++)
	{
		if(carr[i].flag == 1 && carr[i].dir == 0)
		{
			strcpy(mapping, carr[i].M);
			isccDomain(temp, mapping);
			fileHandling(temp);
			strcpy(domain, temp);
			strcpy(str, temp);
//printf("mapping - %s\n", mapping);printf("domain - %s\n", domain);
                        printf("Undirected CF Edge %s %s\n", carr[i].a1.s, carr[i].a2.s);

			supported(carr[i].a1, carr[i].a2, sarr, pws, pcount, cws, ccount, ews, count, domain, mapping);
                        printf("domain %d\n", strcmp(domain, "{  }"));
                        // if Du − Du' \neq \emptyset then
			if(strcmp(str, domain) != 0)
			{
				carr[i].dir = 1;
                                /*Mapping intersection using ISCC functions*/
				isccInterSub(temp, str, domain, "-");
				fileHandling(temp);

                                /*Extract association rules from the mapping*/
				parse(str, temp, 1);
				//remove0(str);
                               /* Modifying the mapping for the changed domain Du - Du' */
				modify(carr[i].M, mapping, str);
                                
			}
                        // if Du' \neq \emptyset then
			if(strcmp(domain, "{  }") != 0)
			{
                                printf("Mummum\n");
				strcpy(str, "M := ");
				strcat(str, mapping);
				strcat(str, ";M^-1;");
				fileHandling(str);

				//remove0(str);
				strcpy(mapping, str);
				strcpy(temp, domain);

				supported(carr[i].a2, carr[i].a1, sarr, pws, pcount, cws, ccount, ews, count, domain, mapping);
				if(strcmp(domain, "{  }") != 0)
				{
					printf("Non Determinate Program\n");
					exit(0);
				}
				else
				{
					tem.a1 = carr[i].a2;
					tem.a2 = carr[i].a1;
					tem.t = !carr[i].t;
					tem.flag = 1;
					tem.dir = 1;
					
					parse(str, temp, 1);
					modify(tem.M, mapping, str);
					carr[(*ecount)++] = tem;
				}
			}
                        printf("domain %d\n", strcmp(domain, "{  }"));
		}
	}

}
